from pylab import *
import numpy as np
from matplotlib.ticker import ScalarFormatter
import matplotlib.gridspec as gridspec
from mpl_toolkits.axes_grid1.inset_locator import mark_inset
from matplotlib.ticker import AutoMinorLocator
from matplotlib.colors import LogNorm
from matplotlib.patches import Ellipse
rcParams["mathtext.fontset"]='cm'


################################## INPUT #######################################
root  = '/mnt/ceph/users/camels/Results'
f_out = 'different_seed.pdf'

# Pk m
f11 = '%s/Pk/IllustrisTNG/mean_LH_Pk_m_z=0.00.txt'%root
f12 = '%s/Pk/SIMBA/mean_LH_Pk_m_z=0.00.txt'%root
#f12 = '%s/Pk/SIMBA_OLD/mean_Pk_m_z=0.00.txt'%root
f13 = '%s/Pk/IllustrisTNG/extreme_0/Pk_m_z=0.00.txt'%root
f14 = '%s/Pk/IllustrisTNG/extremestellar_0/Pk_m_z=0.00.txt'%root
f15 = '%s/Pk/IllustrisTNG/noFB_0/Pk_m_z=0.00.txt'%root

# Pk g
f21 = '%s/Pk/IllustrisTNG/mean_LH_Pk_g_z=0.00.txt'%root
f22 = '%s/Pk/SIMBA/mean_LH_Pk_g_z=0.00.txt'%root
#f22 = '%s/Pk/SIMBA_OLD/mean_Pk_g_z=0.00.txt'%root
f23 = '%s/Pk/IllustrisTNG/extreme_0/Pk_g_z=0.00.txt'%root
f24 = '%s/Pk/IllustrisTNG/extremestellar_0/Pk_g_z=0.00.txt'%root
f25 = '%s/Pk/IllustrisTNG/noFB_0/Pk_g_z=0.00.txt'%root

# ratio Pk m
f31 = '%s/Pk_ratio/IllustrisTNG/mean_LH_Pk_ratio_m_z=0.00.txt'%root
f32 = '%s/Pk_ratio/SIMBA/mean_LH_Pk_ratio_m_z=0.00.txt'%root
#f32 = '%s/Pk_ratio/SIMBA_OLD/mean_Pk_ratio_m_z=0.00.txt'%root
#f33 = '%s/Pk/IllustrisTNG/extreme_0/Pk_g_z=0.00.txt'%root
#f34 = '%s/Pk/IllustrisTNG/extremestellar_0/Pk_g_z=0.00.txt'%root
#f35 = '%s/Pk/IllustrisTNG/noFB_0/Pk_g_z=0.00.txt'%root

# HMF
f41 = '%s/HMF/IllustrisTNG/mean_LH_mass_function_1.00e+10_1.00e+14_30_z=0.00.txt'%root
f42 = '%s/HMF/SIMBA/mean_LH_mass_function_1.00e+10_1.00e+14_30_z=0.00.txt'%root
#f42 = '%s/HMF/SIMBA_OLD/mean_mass_function_1.00e+10_1.00e+14_30_z=0.00.txt'%root
f43 = '%s/HMF/IllustrisTNG/extreme_0/mass_function_1.00e+10_1.00e+14_30_z=0.00.txt'%root
f44 = '%s/HMF/IllustrisTNG/extremestellar_0/mass_function_1.00e+10_1.00e+14_30_z=0.00.txt'%root
f45 = '%s/HMF/IllustrisTNG/noFB_0/mass_function_1.00e+10_1.00e+14_30_z=0.00.txt'%root

# SFRH
f51 = '%s/SFRH/IllustrisTNG/mean_LH_SFRH_0.00_10.00_10000.txt'%root
f52 = '%s/SFRH/SIMBA/mean_LH_SFRH_0.00_10.00_10000.txt'%root
#f52 = '%s/SFRH/SIMBA_OLD/mean_SFRH_0.00_10.00_10000.txt'%root
f53 = '%s/SFRH/IllustrisTNG/extreme_0/SFRH_0.00_10.00_10000.txt'%root
f54 = '%s/SFRH/IllustrisTNG/extremestellar_0/SFRH_0.00_10.00_10000.txt'%root
f55 = '%s/SFRH/IllustrisTNG/noFB_0/SFRH_0.00_10.00_10000.txt'%root

# SMF
f61 = '%s/SMF/IllustrisTNG/mean_LH_SMF_1.00e+09_5e+11_10_z=0.00.txt'%root
f62 = '%s/SMF/SIMBA/mean_LH_SMF_1.00e+09_5e+11_10_z=0.00.txt'%root
#f62 = '%s/SMF/SIMBA_OLD/mean_SMF_1.00e+09_5e+11_10_z=0.00.txt'%root
f63 = '%s/SMF/IllustrisTNG/extreme_0/SMF_1.00e+09_5e+11_10_z=0.00.txt'%root
f64 = '%s/SMF/IllustrisTNG/extremestellar_0/SMF_1.00e+09_5e+11_10_z=0.00.txt'%root
f65 = '%s/SMF/IllustrisTNG/noFB_0/SMF_1.00e+09_5e+11_10_z=0.00.txt'%root

# baryon fraction
f71 = '%s/baryon_fraction/IllustrisTNG/mean_LH_bf_1.00e+10_1.00e+14_30_z=0.00.txt'%root
f72 = '%s/baryon_fraction/SIMBA/mean_LH_bf_1.00e+10_1.00e+14_30_z=0.00.txt'%root
#f72 = '%s/baryon_fraction/SIMBA_OLD/mean_bf_1.00e+10_1.00e+14_30_z=0.00.txt'%root
f73 = '%s/baryon_fraction/IllustrisTNG/extreme_0/bf_1.00e+10_1.00e+14_30_z=0.00.txt'%root
f74 = '%s/baryon_fraction/IllustrisTNG/extremestellar_0/bf_1.00e+10_1.00e+14_30_z=0.00.txt'%root
f75 = '%s/baryon_fraction/IllustrisTNG/noFB_0/bf_1.00e+10_1.00e+14_30_z=0.00.txt'%root

# halos temperature
f81 = '%s/SO/IllustrisTNG/mean_T_LH_1.00e+12_1.00e+14_10_z=0.00.txt'%root
f82 = '%s/SO/SIMBA/mean_T_LH_1.00e+12_1.00e+14_10_z=0.00.txt'%root
#f82 = '%s/SO/SIMBA_OLD/mean_T_1.00e+12_1.00e+14_10_z=0.00.txt'%root
f83 = '%s/SO/IllustrisTNG/extreme_0/SO_z=0.00.txt'%root
f84 = '%s/SO/IllustrisTNG/extremestellar_0/SO_z=0.00.txt'%root
f85 = '%s/SO/IllustrisTNG/noFB_0/SO_z=0.00.txt'%root

# galaxies radius
f91 = '%s/Radii/IllustrisTNG/mean_R_vs_SM_LH_z=0.00.txt'%root
f92 = '%s/Radii/SIMBA/mean_R_vs_SM_LH_z=0.00.txt'%root
#f92 = '%s/Radii/SIMBA_OLD/mean_R_vs_SM_z=0.00.txt'%root
f93 = '%s/Radii/IllustrisTNG/extreme_0/R_vs_SM_1.00e+09_5e+11_10_z=0.00.txt'%root
f94 = '%s/Radii/IllustrisTNG/extremestellar_0/R_vs_SM_1.00e+09_5e+11_10_z=0.00.txt'%root
f95 = '%s/Radii/IllustrisTNG/noFB_0/R_vs_SM_1.00e+09_5e+11_10_z=0.00.txt'%root

# Black hole masses
f101 = '%s/BH/IllustrisTNG/mean_BH_vs_SM_LH_z=0.00.txt'%root
f102 = '%s/BH/SIMBA/mean_BH_vs_SM_LH_z=0.00.txt'%root
#f102 = '%s/BH/SIMBA_OLD/mean_BH_vs_SM_z=0.00.txt'%root
f103 = '%s/BH/IllustrisTNG/extreme_0/BH_vs_SM_1.00e+09_5e+11_10_z=0.00.txt'%root
f104 = '%s/BH/IllustrisTNG/extremestellar_0/BH_vs_SM_1.00e+09_5e+11_10_z=0.00.txt'%root
f105 = '%s/BH/IllustrisTNG/noFB_0/BH_vs_SM_1.00e+09_5e+11_10_z=0.00.txt'%root

# Vmax
f111 = '%s/Vmax/IllustrisTNG/mean_V_vs_SM_LH_z=0.00.txt'%root
f112 = '%s/Vmax/SIMBA/mean_V_vs_SM_LH_z=0.00.txt'%root
#f112 = '%s/Vmax/SIMBA_OLD/mean_V_vs_SM_z=0.00.txt'%root
f113 = '%s/Vmax/IllustrisTNG/extreme_0/Vmax_vs_SM_1.00e+09_5e+11_10_z=0.00.txt'%root
f114 = '%s/Vmax/IllustrisTNG/extremestellar_0/Vmax_vs_SM_1.00e+09_5e+11_10_z=0.00.txt'%root
f115 = '%s/Vmax/IllustrisTNG/noFB_0/Vmax_vs_SM_1.00e+09_5e+11_10_z=0.00.txt'%root

# SFR
f121 = '%s/SFR/IllustrisTNG/mean_SFR_vs_SM_LH_z=0.00.txt'%root
f122 = '%s/SFR/SIMBA/mean_SFR_vs_SM_LH_z=0.00.txt'%root
#f122 = '%s/SFR/SIMBA_OLD/mean_SFR_vs_SM_z=0.00.txt'%root
f123 = '%s/SFR/IllustrisTNG/extreme_0/SFR_vs_SM_1.00e+09_5e+11_10_z=0.00.txt'%root
f124 = '%s/SFR/IllustrisTNG/extremestellar_0/SFR_vs_SM_1.00e+09_5e+11_10_z=0.00.txt'%root
f125 = '%s/SFR/IllustrisTNG/noFB_0/SFR_vs_SM_1.00e+09_5e+11_10_z=0.00.txt'%root
################################################################################

fig  = figure(figsize=(20,22)) 
gs   = gridspec.GridSpec(4,3)
ax1  = plt.subplot(gs[0])
ax2  = plt.subplot(gs[1])
ax3  = plt.subplot(gs[2])
ax4  = plt.subplot(gs[3])
ax5  = plt.subplot(gs[4])
ax6  = plt.subplot(gs[5])
ax7  = plt.subplot(gs[6])
ax8  = plt.subplot(gs[7])
ax9  = plt.subplot(gs[8])
ax10 = plt.subplot(gs[9])
ax11 = plt.subplot(gs[10])
ax12 = plt.subplot(gs[11])
subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=0.2, hspace=0.2)

c1='blue'
c2='red'

###### Pk m plot ######
ax1.set_xscale('log')
ax1.set_yscale('log')
ax1.set_xlim([0.3,35])
ax1.set_ylim([1e-1,1e3])
ax1.set_xlabel(r'$k\,[h{\rm Mpc}^{-1}]$',fontsize=18)
ax1.set_ylabel(r'$P_{\rm m}(k)\,[h^{-3}{\rm Mpc}^3]$',fontsize=18)

X,Y,dY,dYp,dYm,Ym = np.loadtxt(f11, unpack=True)
ax1.fill_between(X, dYp, dYm, color=c1, alpha=0.6)
p1,=ax1.plot(X,Ym,c=c1,linestyle='-')

X,Y,dY,dYp,dYm,Ym = np.loadtxt(f12, unpack=True)
ax1.fill_between(X, dYp, dYm, color=c2, alpha=0.6)
p2,=ax1.plot(X,Ym,c=c2,linestyle='--')

X,Y = np.loadtxt(f13, unpack=True)
#p3,=ax1.plot(X,Y,c='g',linestyle='-')

X,Y = np.loadtxt(f14, unpack=True)
#p3,=ax1.plot(X,Y,c='g',linestyle='--')

X,Y = np.loadtxt(f15, unpack=True)
#p3,=ax1.plot(X,Y,c='g',linestyle='dotted')
#######################

###### Pk g plot ######
ax2.set_xscale('log')
ax2.set_yscale('log')
ax2.set_xlim([0.3,35])
ax2.set_ylim([1e-2,1e3])
ax2.set_xlabel(r'$k\,[h{\rm Mpc}^{-1}]$',fontsize=18)
ax2.set_ylabel(r'$P_{\rm g}(k)\,[h^{-3}{\rm Mpc}^3]$',fontsize=18)

X,Y,dY,dYp,dYm,Ym = np.loadtxt(f21, unpack=True)
ax2.fill_between(X, dYp, dYm, color=c1, alpha=0.6)
ax2.plot(X,Ym,c=c1,linestyle='-')

X,Y,dY,dYp,dYm,Ym = np.loadtxt(f22, unpack=True)
ax2.fill_between(X, dYp, dYm, color=c2, alpha=0.6)
ax2.plot(X,Ym,c=c2,linestyle='--')

X,Y = np.loadtxt(f23, unpack=True)
#p3,=ax2.plot(X,Y,c='g',linestyle='-')

X,Y = np.loadtxt(f24, unpack=True)
#p3,=ax2.plot(X,Y,c='g',linestyle='--')

X,Y = np.loadtxt(f25, unpack=True)
#p3,=ax2.plot(X,Y,c='g',linestyle='dotted')
#######################

###### ratio Pk m plot ######
ax3.set_xscale('log')
ax3.set_xlim([0.3,35])
ax3.set_ylim([0.5,1.05])
ax3.set_xlabel(r'$k\,[h{\rm Mpc}^{-1}]$',fontsize=18)
ax3.set_ylabel(r'$P_{\rm hydro}(k)/P_{\rm Nbody}(k)$',fontsize=18)

X,Y,dY,dYp,dYm,Ym= np.loadtxt(f31, unpack=True)
ax3.fill_between(X, dYp, dYm, color=c1, alpha=0.6)
ax3.plot(X,Ym,c=c1,linestyle='-')

X,Y,dY,dYp,dYm,Ym = np.loadtxt(f32, unpack=True)
ax3.fill_between(X, dYp, dYm, color=c2, alpha=0.6)
ax3.plot(X,Ym,c=c2,linestyle='--')
#######################

######### HMF #########
ax4.set_xscale('log')
ax4.set_yscale('log')
ax4.set_xlabel(r'$M_{\rm halo}/\Omega_{\rm m}\,[h^{-1}M_\odot]$',fontsize=18)
ax4.set_ylabel(r'${\rm HMF}\,[h^{4}{\rm Mpc}^{-3}M_\odot^{-1}]$',fontsize=18)

X,Y,dY,dYp,dYm,Ym = np.loadtxt(f41, unpack=True)
ax4.fill_between(X, dYp, dYm, color=c1, alpha=0.6)
ax4.plot(X,Ym,c=c1,linestyle='-')

X,Y,dY,dYp,dYm,Ym = np.loadtxt(f42, unpack=True)
ax4.fill_between(X, dYp, dYm, color=c2, alpha=0.6)
ax4.plot(X,Ym,c=c2,linestyle='--')

dumb,X,Y = np.loadtxt(f43, unpack=True)
#p3,=ax4.plot(X,Y,c='g',linestyle='-')

dumb,X,Y = np.loadtxt(f44, unpack=True)
#p3,=ax4.plot(X,Y,c='g',linestyle='--')

dumb,X,Y = np.loadtxt(f45, unpack=True)
#p3,=ax4.plot(X,Y,c='g',linestyle='dotted')
#######################

###### SFRH ######
ax5.set_yscale('log')
ax5.set_xlim([0.0,7.0])
ax5.set_ylim([5e-4,0.2])
ax5.set_xlabel(r'$z$',fontsize=18)
ax5.set_ylabel(r'${\rm SFRD}\,[M_\odot{\rm yr}^{-1}{\rm Mpc}^{-3}]$',fontsize=18)

X,Y,dY,dYp,dYm,Ym = np.loadtxt(f51, unpack=True)
ax5.fill_between(X, dYp, dYm, color=c1, alpha=0.6)
ax5.plot(X,Ym,c=c1,linestyle='-')

X,Y,dY,dYp,dYm,Ym = np.loadtxt(f52, unpack=True)
ax5.fill_between(X, dYp, dYm, color=c2, alpha=0.6)
ax5.plot(X,Ym,c=c2,linestyle='--')

X,Y = np.loadtxt(f53, unpack=True)
#p3,=ax5.plot(X,Y,c='g',linestyle='-')

X,Y = np.loadtxt(f54, unpack=True)
#p3,=ax5.plot(X,Y,c='g',linestyle='--')

X,Y = np.loadtxt(f55, unpack=True)
#p3,=ax5.plot(X,Y,c='g',linestyle='dotted')
#######################

###### SMF ######
ax6.set_xscale('log')
ax6.set_yscale('log')
ax6.set_xlabel(r'$M_*\,[h^{-1}M_\odot]$',fontsize=18)
ax6.set_ylabel(r'${\rm SMF}\,[h^4{\rm Mpc}^{-3}M_\odot^{-1}]$',fontsize=18)

X,Y,dY,dYp,dYm,Ym = np.loadtxt(f61, unpack=True)
ax6.fill_between(X, dYp, dYm, color=c1, alpha=0.6)
ax6.plot(X,Ym,c=c1,linestyle='-')

X,Y,dY,dYp,dYm,Ym = np.loadtxt(f62, unpack=True)
ax6.fill_between(X, dYp, dYm, color=c2, alpha=0.6)
ax6.plot(X,Ym,c=c2,linestyle='--')

X,Y = np.loadtxt(f63, unpack=True)
#p3,=ax6.plot(X,Y,c='g',linestyle='-')

X,Y = np.loadtxt(f64, unpack=True)
#p3,=ax6.plot(X,Y,c='g',linestyle='--')

X,Y = np.loadtxt(f65, unpack=True)
#p3,=ax6.plot(X,Y,c='g',linestyle='dotted')
#######################

###### baryon fraction ######
ax7.set_xscale('log')
ax7.set_xlabel(r'$M_{\rm halo}/\Omega_{\rm m}\,[h^{-1}M_\odot]$',fontsize=18)
ax7.set_ylabel(r'$M_{\rm b}/M_{\rm halo}/(\Omega_{\rm b}/\Omega_{\rm m})$',fontsize=18)

X,Y,dY,dYp,dYm,Ym = np.loadtxt(f71, unpack=True)
ax7.fill_between(X, dYp, dYm, color=c1, alpha=0.6)
ax7.plot(X,Ym,c=c1,linestyle='-')

X,Y,dY,dYp,dYm,Ym = np.loadtxt(f72, unpack=True)
ax7.fill_between(X, dYp, dYm, color=c2, alpha=0.6)
ax7.plot(X,Ym,c=c2,linestyle='--')

X,Y = np.loadtxt(f73, unpack=True)
#p3,=ax7.plot(X,Y,c='g',linestyle='-')

X,Y = np.loadtxt(f74, unpack=True)
#p3,=ax7.plot(X,Y,c='g',linestyle='--')

X,Y = np.loadtxt(f75, unpack=True)
#p3,=ax7.plot(X,Y,c='g',linestyle='dotted')
#######################

######## halos temperature #######
ax8.set_xscale('log')
ax8.set_yscale('log')
ax8.set_xlabel(r'$M_{\rm halo}\,[h^{-1}M_\odot]$',fontsize=18)
ax8.set_ylabel(r'$T_{\rm halo}\,[K]$',fontsize=18)

#X,Y,dY,N = np.loadtxt(f81, unpack=True)
#ax8.fill_between(X, Y+dY, Y-dY, color='blue')
X,Y,dY,dYp,dYm,Ym = np.loadtxt(f81, unpack=True)
ax8.fill_between(X, dYp, dYm, color=c1, alpha=0.6)
ax8.plot(X,Ym,c=c1,linestyle='-')

#X,Y,dY,N = np.loadtxt(f82, unpack=True)
#ax8.fill_between(X, Y+dY, Y-dY, color='red', alpha=0.6)
X,Y,dY,dYp,dYm,Ym = np.loadtxt(f82, unpack=True)
ax8.fill_between(X, dYp, dYm, color=c2, alpha=0.6)
ax8.plot(X,Ym,c=c2,linestyle='--')

#data = np.loadtxt(f83)
#ax8.plot(data[:,0], data[:,9],c='k',linestyle='--')

#data = np.loadtxt(f84)
#ax8.plot(data[:,0], data[:,9],c='k',linestyle='--')

#data = np.loadtxt(f85)
#ax8.plot(data[:,0], data[:,9],c='k',linestyle='--')
##################################

############# Radii ##############
ax9.set_xscale('log')
ax9.set_yscale('log')
ax9.set_xlabel(r'$M_*\,[h^{-1}M_\odot]$',fontsize=18)
ax9.set_ylabel(r'$R_{1/2}\,[h^{-1}{\rm kpc}]$',fontsize=18)

#X,Y,dY,N = np.loadtxt(f91, unpack=True)
#ax9.fill_between(X, Y+dY, Y-dY, color='blue')
X,Y,dY,dYp,dYm,Ym = np.loadtxt(f91, unpack=True)
ax9.fill_between(X, dYp, dYm, color=c1, alpha=0.6)
ax9.plot(X,Ym,c=c1,linestyle='-')

#X,Y,dY,N = np.loadtxt(f92, unpack=True)
#ax9.fill_between(X, Y+dY, Y-dY, color='red', alpha=0.6)
X,Y,dY,dYp,dYm,Ym = np.loadtxt(f92, unpack=True)
ax9.fill_between(X, dYp, dYm, color=c2, alpha=0.6)
ax9.plot(X,Ym,c=c2,linestyle='--')

X,Y,dY = np.loadtxt(f93, unpack=True)
#p3,=ax9.plot(X,Y,c='g',linestyle='-')

X,Y,dY = np.loadtxt(f94, unpack=True)
#p3,=ax9.plot(X,Y,c='g',linestyle='--')

X,Y,dY = np.loadtxt(f95, unpack=True)
#p3,=ax9.plot(X,Y,c='g',linestyle='dotted')
##################################

############# BH masses ##############
ax10.set_xscale('log')
ax10.set_yscale('log')
ax10.set_xlabel(r'$M_*\,[h^{-1}M_\odot]$',fontsize=18)
ax10.set_ylabel(r'$M_{\rm black-holes}\,[h^{-1}M_\odot]$',fontsize=18)

#X,Y,dY,N = np.loadtxt(f101, unpack=True)
#ax10.fill_between(X, Y+dY, Y-dY, color='blue')
X,Y,dY,dYp,dYm,Ym = np.loadtxt(f101, unpack=True)
ax10.fill_between(X, dYp, dYm, color=c1, alpha=0.6)
ax10.plot(X,Ym,c=c1,linestyle='-')

#X,Y,dY,N = np.loadtxt(f102, unpack=True)
#ax10.fill_between(X, Y+dY, Y-dY, color='red', alpha=0.6)
X,Y,dY,dYp,dYm,Ym = np.loadtxt(f102, unpack=True)
ax10.fill_between(X, dYp, dYm, color=c2, alpha=0.6)
ax10.plot(X,Ym,c=c2,linestyle='--')

X,Y,dY = np.loadtxt(f103, unpack=True)
#p3,=ax10.plot(X,Y,c='g',linestyle='-')

X,Y,dY = np.loadtxt(f104, unpack=True)
#p3,=ax10.plot(X,Y,c='g',linestyle='--')

X,Y,dY = np.loadtxt(f105, unpack=True)
#p3,=ax10.plot(X,Y,c='g',linestyle='dotted')
##################################

############# Vmax ##############
ax11.set_xscale('log')
ax11.set_yscale('log')
ax11.set_xlabel(r'$M_*\,[h^{-1}M_\odot]$',fontsize=18)
ax11.set_ylabel(r'$V_{\rm max}\,[{\rm km/s}]$',fontsize=18)

#X,Y,dY,N = np.loadtxt(f111, unpack=True)
#ax11.fill_between(X, Y+dY, Y-dY, color='blue')
X,Y,dY,dYp,dYm,Ym = np.loadtxt(f111, unpack=True)
ax11.fill_between(X, dYp, dYm, color=c1, alpha=0.6)
ax11.plot(X,Ym,c=c1,linestyle='-')

#X,Y,dY,N = np.loadtxt(f112, unpack=True)
#ax11.fill_between(X, Y+dY, Y-dY, color='red', alpha=0.6)
X,Y,dY,dYp,dYm,Ym = np.loadtxt(f112, unpack=True)
ax11.fill_between(X, dYp, dYm, color=c2, alpha=0.6)
ax11.plot(X,Ym,c=c2,linestyle='--')

X,Y,dY = np.loadtxt(f113, unpack=True)
#p3,=ax11.plot(X,Y,c='g',linestyle='-')

X,Y,dY = np.loadtxt(f114, unpack=True)
#p3,=ax11.plot(X,Y,c='g',linestyle='--')

X,Y,dY = np.loadtxt(f115, unpack=True)
#p3,=ax11.plot(X,Y,c='g',linestyle='dotted')
##################################

############# SFR ##############
ax12.set_xscale('log')
ax12.set_yscale('log')
ax12.set_xlabel(r'$M_*\,[h^{-1}M_\odot]$',fontsize=18)
ax12.set_ylabel(r'${\rm SFR}\,[M_\odot{\rm yr}^{-1}]$',fontsize=18)

#X,Y,dY,N = np.loadtxt(f121, unpack=True)
#ax12.fill_between(X, Y+dY, Y-dY, color='blue')
X,Y,dY,dYp,dYm,Ym = np.loadtxt(f121, unpack=True)
ax12.fill_between(X, dYp, dYm, color=c1, alpha=0.6)
ax12.plot(X,Ym,c=c1,linestyle='-')

#X,Y,dY,N = np.loadtxt(f122, unpack=True)
#ax12.fill_between(X, Y+dY, Y-dY, color='red', alpha=0.6)
X,Y,dY,dYp,dYm,Ym = np.loadtxt(f122, unpack=True)
ax12.fill_between(X, dYp, dYm, color=c2, alpha=0.6)
ax12.plot(X,Ym,c=c2,linestyle='--')

X,Y,dY = np.loadtxt(f123, unpack=True)
#p3,=ax12.plot(X,Y,c='g',linestyle='-')

X,Y,dY = np.loadtxt(f124, unpack=True)
#p3,=ax12.plot(X,Y,c='g',linestyle='--')

X,Y,dY = np.loadtxt(f125, unpack=True)
#p3,=ax12.plot(X,Y,c='g',linestyle='dotted')
##################################

ax1.text(0.05,0.05, r"$I$",    fontsize=22, color='k',transform=ax1.transAxes)
ax2.text(0.05,0.05, r"$II$",   fontsize=22, color='k',transform=ax2.transAxes)
ax3.text(0.05,0.05, r"$III$",  fontsize=22, color='k',transform=ax3.transAxes)
ax4.text(0.05,0.05, r"$IV$",   fontsize=22, color='k',transform=ax4.transAxes)
ax5.text(0.05,0.05, r"$V$",    fontsize=22, color='k',transform=ax5.transAxes)
ax6.text(0.05,0.05, r"$VI$",   fontsize=22, color='k',transform=ax6.transAxes)
ax7.text(0.05,0.9,  r"$VII$",  fontsize=22, color='k',transform=ax7.transAxes)
ax8.text(0.05,0.9,  r"$VIII$", fontsize=22, color='k',transform=ax8.transAxes)
ax9.text(0.05,0.9,  r"$IX$",   fontsize=22, color='k',transform=ax9.transAxes)
ax10.text(0.05,0.9, r"$X$",    fontsize=22, color='k',transform=ax10.transAxes)
ax11.text(0.05,0.9, r"$XI$",   fontsize=22, color='k',transform=ax11.transAxes)
ax12.text(0.05,0.9, r"$XII$",  fontsize=22, color='k',transform=ax12.transAxes)

#legend
ax1.legend([p1,p2],
           [r"${\rm IllustrisTNG}$",
            r"${\rm SIMBA}$"],
           loc=0,prop={'size':18},ncol=1,frameon=True)

#ax1.set_title(r'$\sum m_\nu=0.0\/{\rm eV}$',position=(0.5,1.02),size=18)
#title('About as simple as it gets, folks')
#suptitle('About as simple as it gets, folks')  #for title with several panels
#grid(True)
#show()
savefig(f_out, bbox_inches='tight')
close(fig)








# This script contains different routines to analyze the output of the CAMEL simulations
import numpy as np
import sys,os,h5py,time
import readgadget
import MAS_library as MASL
import units_library as UL
import sorting_library as SL
import Pk_library as PKL
import HI_library as HIL
import scipy.spatial as SS

# define constants here
rho_crit   = UL.units().rho_crit #h^2 Msun/Mpc^3
BOLTZMANN  = 1.38065e-16         #erg/K - NIST 2010
PROTONMASS = 1.67262178e-24      #gram  - NIST 2010

########################################################################################
# This function computes the distance of each gas particle to its Nneigh nearest 
# neighbourgs
# pos1 -------> positions of the particles
# pos2 -------> positions of the tracers (can be the same as pos1)
# k ----------> compute distance to Nneigh closest neighbourghs
# BoxSize ----> for non-periodic boundary conditions set this to None
# treads -----> number of openmp threads
# verbose ----> whether to print some information on the progress
def KDTree_distance(pos1, pos2, k, BoxSize=None, threads=1, verbose=True):

    # construct kdtree of the particles
    start  = time.time()
    kdtree = SS.cKDTree(pos1, leafsize=16, boxsize=BoxSize)
    if verbose:  print('Time to build KDTree = %.3f seconds'%(time.time()-start))

    # find nearest neighbors of the tracer particles
    start = time.time()
    dist, indexes = kdtree.query(pos2, k, n_jobs=threads)
    if verbose:  print('Time to find k-neighbors = %.3f seconds'%(time.time()-start))

    # return the distance of each particle to its farthest neighborgh
    return dist[:,-1]
########################################################################################

########################################################################################
# This routine returns the suffix of the considered ptype
def Pk_suffix(ptype):
    if   ptype==[0]:        return 'g'
    elif ptype==[1]:        return 'c'
    elif ptype==[4]:        return 's'
    elif ptype==[5]:        return 'bh'
    elif ptype==[0,1,4,5]:  return 'm'
    else:  raise Exception('no label found for ptype')
########################################################################################

########################################################################################
# This routine computes the power spectrum of a snapshot
# snapshot -----> snapshot name
# grid ---------> size of the grid to compute Pk
# MAS ----------> Mass Assignment Scheme (typically 'CIC')
# threads ------> number of OPENMP threads used to compute Pk
# ptype --------> particle type: [0]-gas, [1]-DM, [4]-stars, [5]-black holes
# root_out -----> folder where to save the power spectrum
def compute_Pk(snapshot, grid, MAS, threads, ptype, root_out):

    # read header
    if not(os.path.exists(snapshot)):  return 0
    head     = readgadget.header(snapshot)
    BoxSize  = head.boxsize/1e3  #Mpc/h  
    Nall     = head.nall         #Total number of particles
    Masses   = head.massarr*1e10 #Masses of the particles in Msun/h                    
    Omega_m  = head.omega_m
    Omega_l  = head.omega_l
    redshift = head.redshift
    Hubble   = 100.0*np.sqrt(Omega_m*(1.0+redshift)**3+Omega_l)#km/s/(Mpc/h)
    h        = head.hubble
    Ntot     = np.sum(Nall[ptype], dtype=np.int64)

    # get the name of the output file
    fout = '%s/Pk_%s_z=%.2f.txt'%(root_out, Pk_suffix(ptype), redshift)
    if os.path.exists(fout):  return 0

    # define the arrays containing the number positions and masses of the particles
    pos  = np.zeros((Ntot,3), dtype=np.float32)
    mass = np.zeros(Ntot,     dtype=np.float32)

    # read data for the different particle types
    f = h5py.File(snapshot, 'r');  offset = 0
    for pt in ptype:
        # sometimes there are not black-holes or stars...
        if 'PartType%d'%pt not in f.keys():  continue

        # read positions
        pos_pt  = f['PartType%d/Coordinates'%pt][:]/1e3  #Mpc/h
        if pos_pt.dtype==np.float64:  pos_pt = pos_pt.astype(np.float32)

        # read masses
        if 'PartType%d/Masses'%pt in f:
            mass_pt = f['PartType%d/Masses'%pt][:]*1e10                    #Msun/h
        else:
            mass_pt = np.ones(pos_pt.shape[0], dtype=np.float32)*Masses[1] #Msun/h

        # fill pos and mass arrays
        length  = len(pos_pt)
        pos[offset:offset+length]  = pos_pt
        mass[offset:offset+length] = mass_pt
        offset += length
    f.close()
    if offset!=Ntot:  raise Exception('Not all particles counted')

    # calculate density field
    delta = np.zeros((grid,grid,grid), dtype=np.float32)
    MASL.MA(pos, delta, BoxSize, MAS, W=mass)
    delta /= np.mean(delta, dtype=np.float64);  delta -= 1.0 

    # compute Pk and save results to file
    axis = 0
    Pk = PKL.Pk(delta, BoxSize, axis, MAS, threads)
    np.savetxt(fout, np.transpose([Pk.k3D, Pk.Pk[:,0]]), delimiter='\t')
########################################################################################

########################################################################################
# This routine computes the Pk of the ICs
# snapshot -----> snapshot name
# grid ---------> size of the grid to compute Pk
# MAS ----------> Mass Assignment Scheme (typically 'CIC')
# threads ------> number of OPENMP threads used to compute Pk
# ptype --------> particle type: [0]-gas, [1]-DM, [4]-stars, [5]-black holes
# root_out -----> folder where to save the power spectrum
def compute_Pk_ICs(snapshot, grid, MAS, threads, ptype, root_out):

    if not(os.path.exists(snapshot)) and not(os.path.exists(snapshot+'.0')):  return 0

    # read header
    head     = readgadget.header(snapshot)
    BoxSize  = head.boxsize/1e3  #Mpc/h  
    redshift = head.redshift

    # get the name of the file
    fout = '%s/Pk_%s_z=%.2f.txt'%(root_out, Pk_suffix(ptype), redshift)
    if os.path.exists(fout):  return 0
    
    # compute overdensity field
    do_RSD, axis = False, 0
    delta = MASL.density_field_gadget(snapshot, ptype, grid, MAS, do_RSD, axis)
    delta /= np.mean(delta, dtype=np.float64);  delta -= 1.0

    # compute Pk and save results to file
    Pk = PKL.Pk(delta, BoxSize, axis, MAS, threads)
    np.savetxt(fout, np.transpose([Pk.k3D, Pk.Pk[:,0]]), delimiter='\t')
########################################################################################

########################################################################################
# This routine computes Pk_hydro / Pk_Nbody
# root --------> main CAMEL folder
# sim ---------> simulation type 'IllustrisTNG' or 'SIMBA'
# i -----------> realization number
# snapnum -----> snapnum of the snapshot. Needed to get the redshift
# root_out ----> output folder
def compute_Pk_ratio(root, sim, i, snapnum, root_out):

    # find the number of N-body counterpart
    #if   i in map(str,np.arange(1522,1566)):                       j = 1505
    #elif i in ['1505_0', '1505_1', '1505_2', '1505_3', '1505_4']:  j = 1505
    #elif i in ['0_test', '0_clean0', '0_clean1']:                  j = 0
    #else:                                                          j = i

    #if i in ['1505_0', '1505_1', '1505_2', '1505_3', '1505_4', '1505_5', '1505_6',
    #         '1505_7', '1505_8', '1505_9', '1505_10', '1505_11', '1505_12', '1505_13',
    #         '1505_14', '1505_15', '1505_16', '1505_17', '1505_18', '1505_19', 
    #         '1505_20', '1505_21', '1505_22', '1505_23', '1505_24', '1505_25', 
    #         '1505_26'] and sim=='IllustrisTNG':
    #    sim2='SIMBA'
    #else:  sim2=sim

    # get the names of the snapshots
    #snapshot1 = '%s/Sims/%s/%s/snap_%03d.hdf5'%(root,sim,i,snapnum)
    #snapshot2 = '%s/Sims/%s_DM/%s/snap_%03d.hdf5'%(root,sim2,j,snapnum)
    snapshot1 = '%s/Sims/%s/%s/snap_%03d.hdf5'%(root,sim,i,snapnum)
    snapshot2 = '%s/Sims/%s_DM/%s/snap_%03d.hdf5'%(root,sim,i,snapnum)
    if not(os.path.exists(snapshot1)) or not(os.path.exists(snapshot2)): return 0

    # read the redshifts of the snapshots
    redshift1 = (readgadget.header(snapshot1)).redshift
    redshift2 = (readgadget.header(snapshot2)).redshift

    # get the name of the output file
    fout = '%s/Pk_ratio_m_z=%.2f.txt'%(root_out, redshift2)
    if os.path.exists(fout):  return 0

    # get the name of the power spectra
    f_hydro = '%s/Results/Pk/%s/%s/Pk_m_z=%.2f.txt'%(root,sim,i,redshift1)
    f_nbody = '%s/Results/Pk/%s_DM/%s/Pk_m_z=%.2f.txt'%(root,sim,i,redshift2)
    #f_nbody = '%s/Results/Pk/%s_DM/%s/Pk_m_z=%.2f.txt'%(root,sim,j,redshift2)
    if not(os.path.exists(f_hydro)) or not(os.path.exists(f_nbody)):  return 0

    # read power spectra and save results to file
    k1, Pk_hydro = np.loadtxt(f_hydro, unpack=True)
    k2, Pk_nbody = np.loadtxt(f_nbody, unpack=True)
    if np.any(k1!=k2):  raise Exception('k-values differ!')
    np.savetxt(fout, np.transpose([k1,Pk_hydro/Pk_nbody]))
########################################################################################

########################################################################################
# This routine computes the baryon fraction in FoF halos
# RMmin -------> minimum reduced mass (mass/Omega_m) in the halo
# RMmax -------> maximum reduced mass (mass/Omega_m) in the halo
# bins --------> baryon fraction is computed in log from Nmin to Nmax using # bins
# f_subfind ---> name of the subfind file
# snapshot ----> snapshot name (to read CDM particle masses)
# root_out ----> folder where to write the results
def baryon_fraction_FoF(RMmin, RMmax, bins, f_subfind, snapshot, root_out):

    # check if subfind file exists
    if not(os.path.exists(f_subfind)):  return 0

    # read header and get masses of CDM particles
    head     = readgadget.header(snapshot)
    BoxSize  = head.boxsize/1e3  #Mpc/h  
    Masses   = head.massarr*1e10 #Masses of the particles in Msun/h       
    Nall     = head.nall         #Total number of particles
    Om       = head.omega_m
    redshift = head.redshift
    Mc       = (Om - 0.049)*BoxSize**3*rho_crit/Nall[1] #mass of a CDM particle

    # get the name of the output file
    fout = '%s/bf_%.2e_%.2e_%d_z=%.2f.txt'%(root_out,RMmin, RMmax, bins, redshift)
    if os.path.exists(fout):  return 0

    # read halo masses
    f              = h5py.File(f_subfind, 'r')
    halo_mass      = f['Group/GroupMass'][:]*1e10
    halo_mass_type = f['Group/GroupMassType'][:]*1e10
    halo_part_type = f['Group/GroupLenType'][:]   #number of particles in each halo
    f.close()

    # take only halos with more than 50 CDM particles
    indexes = np.where(halo_part_type[:,1]>50)[0]
    halo_mass = halo_mass[indexes]
    halo_mass_type = halo_mass_type[indexes]

    # define the bins with the number of CDM particles and the intervals mean
    RM_bins = np.logspace(np.log10(RMmin), np.log10(RMmax), bins+1)
    RM_mean = 10**(0.5*(np.log10(RM_bins[1:]) + np.log10(RM_bins[:-1])))

    # compute baryon fraction in units of cosmic fraction
    fraction = ((halo_mass_type[:,0] + halo_mass_type[:,4] + halo_mass_type[:,5])/halo_mass) / (0.049/Om)

    # take bins in halo mass / Omega_m. Compute average baryon fraction
    mean_fraction = np.histogram(halo_mass/Om, RM_bins, weights=fraction)[0]
    Number        = np.histogram(halo_mass/Om, RM_bins)[0]
    Number[np.where(Number==0)] = 1.0
    mean_fraction = mean_fraction/Number

    # save results to file
    np.savetxt(fout, np.transpose([RM_mean, mean_fraction]))
########################################################################################

########################################################################################
# This routine computes the baryon fraction in FoF halos
# RMmin -------> minimum reduced mass (mass/Omega_m) in the halo
# RMmax -------> maximum reduced mass (mass/Omega_m) in the halo
# bins --------> baryon fraction is computed in log from Nmin to Nmax using # bins
# f_SO --------> name of the SO file
# snapshot ----> snapshot name (to read CDM particle masses)
# root_out ----> folder where to write the results
def baryon_fraction_SO(RMmin, RMmax, bins, f_SO, snapshot, root_out):

    # check if SO file exists
    if not(os.path.exists(f_SO)):  return 0

    # read header and get masses of CDM particles
    head     = readgadget.header(snapshot)
    BoxSize  = head.boxsize/1e3  #Mpc/h  
    Om       = head.omega_m
    redshift = head.redshift

    # get the name of the output file
    fout1 = '%s/bf_SO_%.2e_%.2e_%d_z=%.2f.txt'%(root_out, RMmin, RMmax, bins, redshift)
    fout2 = '%s/gf_SO_%.2e_%.2e_%d_z=%.2f.txt'%(root_out, RMmin, RMmax, bins, redshift)
    if os.path.exists(fout1) and os.path.exists(fout2):  return 0

    # read halo masses
    data = np.loadtxt(f_SO, unpack=False)
    halo_mass = data[:,0]
    Mg        = data[:,5]
    Mc        = data[:,6]
    Ms        = data[:,7]
    Mbh       = data[:,8]
    Nc        = data[:,11]

    # take only halos with more than 50 CDM particles
    indexes   = np.where(Nc>50)[0]
    halo_mass = halo_mass[indexes]
    Mg        = Mg[indexes]
    Mc        = Mc[indexes]
    Ms        = Ms[indexes]
    Mbh       = Mbh[indexes]

    # define the bins with the number of CDM particles and the intervals mean
    RM_bins = np.logspace(np.log10(RMmin), np.log10(RMmax), bins+1)
    RM_mean = 10**(0.5*(np.log10(RM_bins[1:]) + np.log10(RM_bins[:-1])))

    # compute baryon fraction in units of cosmic fraction
    fraction1 = ((Mg + Ms + Mbh)/halo_mass) / (0.049/Om)
    fraction2 = (Mg/halo_mass) / (0.049/Om)

    # take bins in halo mass / Omega_m. Compute average baryon fraction
    mean_fraction1 = np.histogram(halo_mass/Om, RM_bins, weights=fraction1)[0]
    mean_fraction2 = np.histogram(halo_mass/Om, RM_bins, weights=fraction2)[0]
    Number         = np.histogram(halo_mass/Om, RM_bins)[0]
    Number[np.where(Number==0)] = 1.0
    mean_fraction1 = mean_fraction1/Number
    mean_fraction2 = mean_fraction2/Number

    # save results to file
    np.savetxt(fout1, np.transpose([RM_mean, mean_fraction1]))
    np.savetxt(fout2, np.transpose([RM_mean, mean_fraction2]))
########################################################################################

########################################################################################
# This routine computes the mean and std of the halo mass function
# We use reduced mass instead of traditional mass because Omega_m changes a lot in the
# CAMEL simulations. Results are less sensitive to Omega_m using the reduced mass
# RMmin -------> minimum reduced mass (mass/Omega_m) in the halo
# RMmax -------> maximum reduced mass (mass/Omega_m) in the halo
# bins --------> baryon fraction is computed in log from Nmin to Nmax using # bins
# f_subfind ---> name of the subfind file
# snapshot ----> snapshot name (to read CDM particle masses)
# root_out ----> folder where to write the results
def halo_mass_function(RMmin, RMmax, bins, f_subfind, snapshot, root_out):

    # check if subfind file exists
    if not(os.path.exists(f_subfind)):  return 0

    # read header and get masses of CDM particles
    head     = readgadget.header(snapshot)
    BoxSize  = head.boxsize/1e3  #Mpc/h  
    Masses   = head.massarr*1e10 #Masses of the particles in Msun/h                    
    Nall     = head.nall         #Total number of particles
    Om       = head.omega_m
    redshift = head.redshift

    # get the name of the output file
    fout = '%s/mass_function_%.2e_%.2e_%d_z=%.2f.txt'%(root_out,RMmin,RMmax,bins,redshift)
    if os.path.exists(fout):  return 0

    # read halo masses
    f              = h5py.File(f_subfind, 'r')
    halo_mass      = f['Group/GroupMass'][:]*1e10 #Msun/h
    halo_part_type = f['Group/GroupLenType'][:]   #number of particles in each halo
    f.close()

    # take only halos with more than 50 CDM particles
    indexes   = np.where(halo_part_type[:,1]>50)[0]
    halo_mass = halo_mass[indexes]

    # define the arrays with the reduced mass (mass/Omega_m) bins, mean and width
    RM_bins = np.logspace(np.log10(RMmin), np.log10(RMmax), bins+1)
    RM_mean = 10**(0.5*(np.log10(RM_bins[1:]) + np.log10(RM_bins[:-1])))
    dRM     = RM_bins[1:] - RM_bins[:-1]

    # compute halo mass function and save results to file
    HMF = np.histogram(halo_mass/Om, RM_bins)[0]
    HMF = HMF/(BoxSize**3*dRM*Om)
    np.savetxt(fout, np.transpose([RM_mean*Om, RM_mean, HMF]))
########################################################################################

########################################################################################
# This routine reads the SFRH from the sims and produce a lighter file with the SFRH
# from z_min to z_max
# f_sfr ------> file with the star-formation rate history
# z_min ------> minimum redshift
# z_max ------> maximum redshift
# bins -------> number of bins to use
# BoxSize ----> Size of the simulation box
# root_out ---> output folder
def star_formation_rate_history(f_sfrh, z_min, z_max, bins, BoxSize, root_out):

    h = 0.6711

    # check if files exists
    fout = '%s/SFRH_%.2f_%.2f_%d.txt'%(root_out, z_min, z_max, bins)
    if os.path.exists(fout):  return 0
    if not(os.path.exists(f_sfrh)):  return 0

    # define the z-bins
    bins_z = np.linspace(z_min, z_max, bins)

    # read SFRH and interpolate and save results to file
    data = np.loadtxt(f_sfrh);  z, SFR = 1.0/data[:,0]-1.0, data[:,2]/(BoxSize/h)**3
    indexes = np.argsort(z);  z = z[indexes];  SFR = SFR[indexes]
    SFRH = np.interp(bins_z, z, SFR)
    #SFRH = np.interp(bins_z, z[::-1], SFR[::-1])
    np.savetxt(fout, np.transpose([bins_z, SFRH]))

# This routine computes the mean and std of the considered realizations
# sim ------------> simulation considered: 'SIMBA' or 'IllustrisTNG'
# z_min ----------> minimum redshift
# z_max ----------> maximum redshift
# bins -----------> number of bins
# root -----------> folder where data is
# realizations ---> array with the realizations to use for the mean and std
# fout -----------> name of output file
def mean_SFRH(sim, z_min, z_max, bins, root, realizations, fout, verbose=False):
    
    # define the array hosting all data
    SFRH = np.zeros((len(realizations),bins), dtype=np.float64)

    # read all SFRH and fill the array
    count = 0
    for i in realizations:
        fin = '%s/%s/%d/SFRH_%.2f_%.2f_%d.txt'%(root,sim,i,z_min,z_max,bins)
        if not(os.path.exists(fin)):  continue
        z, SFRH[count] = np.loadtxt(fin, unpack=True)
        count += 1
    if verbose:  print('Found %d realizations for %s'%(count,sim))
    SFRH = SFRH[:count]
    
    # save mean and std to file
    np.savetxt(fout, np.transpose([z, np.mean(SFRH, axis=0), np.std(SFRH, axis=0),
                                   np.percentile(SFRH,84,axis=0), 
                                   np.percentile(SFRH,16,axis=0)]))
########################################################################################

########################################################################################
# This routine computes the stellar mass function of a subfind catalogue
# SMmin -------> minimum value of the stellar mass; Msun/h
# SMmax -------> maximum value of the stellar mass; Msun/h
# bins --------> number of bins to compute the stellar mass function
# f_subfind ---> subfind catalogue
# root_out ----> output folder
def stellar_mass_function(SMmin, SMmax, bins, f_subfind, root_out):

    if not(os.path.exists(f_subfind)):  return 0

    # read stellar masses
    f = h5py.File(f_subfind, 'r')
    BoxSize  = f['Header'].attrs[u'BoxSize']/1e3 #Mpc/h
    redshift = f['Header'].attrs[u'Redshift']
    SM       = f['Subhalo/SubhaloMassType'][:]*1e10 #Msun/h
    SM       = SM[:,4]
    f.close()

    # get the name of the output file
    fout = '%s/SMF_%.2e_%2.e_%d_z=%.2f.txt'%(root_out, SMmin, SMmax, bins, redshift)
    if os.path.exists(fout):  return 0

    # define bins, mean and width of the stellar mass function
    bins_SM = np.logspace(np.log10(SMmin), np.log10(SMmax), bins+1)
    mean_SM = 0.5*(bins_SM[1:] + bins_SM[:-1])
    dSM     = bins_SM[1:] - bins_SM[:-1]

    # compute stellar mass function and save results to file
    SMF = np.histogram(SM, bins_SM)[0]
    SMF = SMF/(BoxSize**3*dSM)
    np.savetxt(fout, np.transpose([mean_SM, SMF]))
########################################################################################

########################################################################################
# This routine computes the average of different galaxy properties vs stellar mass
# SMmin -------> minimum value of the stellar mass; Msun/h
# SMmax -------> maximum value of the stellar mass; Msun/h
# bins --------> number of bins to compute the stellar mass function
# f_subfind ---> subfind catalogue
# root_out ----> output folder
# sim ---------> sim type 'SIMBA', 'IllustrisTNG'
# i -----------> name of the realization: e.g. '0' or '1505_0'
def properties_vs_SM(SMmin, SMmax, bins, f_subfind, root_out, sim, i):

    if not(os.path.exists(f_subfind)):  return 0

    # read stellar masses, 
    f = h5py.File(f_subfind, 'r')
    BoxSize  = f['Header'].attrs[u'BoxSize']/1e3 #Mpc/h
    redshift = f['Header'].attrs[u'Redshift']
    SM   = f['Subhalo/SubhaloMassType'][:][:,4]*1e10   #Msun/h
    Rp   = f['Subhalo/SubhaloHalfmassRadType'][:][:,4] #kpc/h
    BHp  = f['Subhalo/SubhaloBHMass'][:]*1e10          #Msun/h
    Vp   = f['Subhalo/SubhaloVmax'][:]                 #km/s
    SFRp = f['Subhalo/SubhaloSFR'][:]                  #Msun/yr
    f.close()

    # define bins, mean and width of the stellar mass function
    bins_SM = np.logspace(np.log10(SMmin), np.log10(SMmax), bins+1)
    mean_SM = 0.5*(bins_SM[1:] + bins_SM[:-1])
    dSM     = bins_SM[1:] - bins_SM[:-1]

    # number of galaxies in each stellar bin
    N = np.histogram(SM, bins_SM)[0]

    ##### Radii #####
    R = np.histogram(SM, bins_SM, weights=Rp)[0];  
    indexes = np.where(R!=0.0)[0];  R[indexes] = R[indexes]/N[indexes]
    root_out2 = '%s/Radii/%s/%s'%(root_out, sim, i)
    if not(os.path.exists(root_out2)):  os.makedirs(root_out2)
    fout = '%s/R_vs_SM_%.2e_%2.e_%d_z=%.2f.txt'%(root_out2,SMmin,SMmax,bins,redshift)
    np.savetxt(fout, np.transpose([mean_SM, R, N]))

    ##### BH mass #####
    BH = np.histogram(SM, bins_SM, weights=BHp)[0]
    indexes = np.where(BH!=0.0)[0];   BH[indexes] = BH[indexes]/N[indexes]
    root_out2 = '%s/BH/%s/%s'%(root_out, sim, i)
    if not(os.path.exists(root_out2)):  os.makedirs(root_out2)
    fout = '%s/BH_vs_SM_%.2e_%2.e_%d_z=%.2f.txt'%(root_out2,SMmin,SMmax,bins,redshift)
    np.savetxt(fout, np.transpose([mean_SM, BH, N]))

    ##### SFR #####
    SFR = np.histogram(SM, bins_SM, weights=SFRp)[0]
    indexes = np.where(SFR!=0.0)[0];   SFR[indexes] = SFR[indexes]/N[indexes]
    root_out2 = '%s/SFR/%s/%s'%(root_out, sim, i)
    if not(os.path.exists(root_out2)):  os.makedirs(root_out2)
    fout = '%s/SFR_vs_SM_%.2e_%2.e_%d_z=%.2f.txt'%(root_out2,SMmin,SMmax,bins,redshift)
    np.savetxt(fout, np.transpose([mean_SM, SFR, N]))

    ##### Vmax #####
    indexes = np.where(Vp!=np.inf)[0]
    V = np.histogram(SM[indexes], bins_SM, weights=Vp[indexes])[0]
    indexes = np.where(V!=0.0)[0];  V[indexes] = V[indexes]/N[indexes]
    root_out2 = '%s/Vmax/%s/%s'%(root_out, sim, i)
    if not(os.path.exists(root_out2)):  os.makedirs(root_out2)
    fout = '%s/Vmax_vs_SM_%.2e_%2.e_%d_z=%.2f.txt'%(root_out2,SMmin,SMmax,bins,redshift)
    np.savetxt(fout, np.transpose([mean_SM, V, N]))
########################################################################################

########################################################################################
# This function computes the temperature of the gas particles of a given snapshot
def temperature(snapshot):

    # read internal energy, electron abundance  and star-formation rate
    f      = h5py.File(snapshot, 'r')
    ne     = f['/PartType0/ElectronAbundance'][:]
    energy = f['/PartType0/InternalEnergy'][:] #(km/s)^2
    f.close()

    # compute the temperature
    yhelium = 0.0789
    T = energy*(1.0 + 4.0*yhelium)/(1.0 + yhelium + ne)*1e10*(2.0/3.0)
    T *= (PROTONMASS/BOLTZMANN)
    return T
########################################################################################

########################################################################################
# This function computes the HI mass of the gas particles of a given snapshot
def HI_mass(snapshot, TREECOOL_file, sim='IllustrisTNG'):

    # read redshift and h
    f = h5py.File(snapshot, 'r')
    redshift = f['Header'].attrs[u'Redshift']
    h        = f['Header'].attrs[u'HubbleParam']

    # read pos, radii, densities, HI/H and masses of gas particles 
    MHI  = f['PartType0/NeutralHydrogenAbundance'][:]
    mass = f['PartType0/Masses'][:]*1e10  #Msun/h
    rho  = f['PartType0/Density'][:]*1e19 #(Msun/h)/(Mpc/h)^3
    SFR  = f['PartType0/StarFormationRate'][:]
    indexes = np.where(SFR>0.0)[0];  del SFR
            
    # find the metallicity of star-forming particles
    if sim in ['IllustrisTNG', 'Astrid']:
        metals = f['PartType0/GFM_Metallicity'][:]
    elif sim=='SIMBA':
        metals = f['PartType0/Metallicity'][:,0] #metallicity
    else:
        raise Exception('Wrong simulation type!!!')
    metals = metals[indexes]/0.0127
    f.close()

    # find densities of star-forming particles: units of h^2 Msun/Mpc^3
    Volume = mass/rho                            #(Mpc/h)^3
    radii  = (Volume/(4.0*np.pi/3.0))**(1.0/3.0) #Mpc/h 
    rho    = rho[indexes]                        #h^2 Msun/Mpc^3
    Volume = Volume[indexes]                     #(Mpc/h)^3

    # find volume and radius of star-forming particles
    radii_SFR  = (Volume/(4.0*np.pi/3.0))**(1.0/3.0) #Mpc/h 
        
    # find HI/H fraction for star-forming particles
    MHI[indexes] = HIL.Rahmati_HI_Illustris(rho, radii_SFR, metals, redshift, h, 
                                            TREECOOL_file, Gamma=None, fac=1, 
                                            correct_H2=True) #HI/H
    MHI *= (0.76*mass)
    return MHI

########################################################################################

########################################################################################
# This function computes the pressure of the gas particles of a given snapshot
def pressure(snapshot):

    gamma = 5.0/3.0

    # read internal energy, electron abundance  and star-formation rate
    f   = h5py.File(snapshot, 'r')
    rho = f['/PartType0/Density'][:]*1e10   #(Msun/h)/(kpc/h)^3
    U   = f['/PartType0/InternalEnergy'][:] #(km/s)^2
    f.close()

    P = (gamma-1.0)*U*rho

    return P  #units are (Msun/h)*(km/s)^2/(kpc/h)^3
########################################################################################

########################################################################################
# This function computes the pressure of the gas particles of a given snapshot
def electron_density(snapshot):

    m_proton = 1.6726e-27    #kg
    Msun     = 1.99e30       #kg
    kpc      = 3.0857e21     #cm

    # read internal energy, electron abundance  and star-formation rate
    f    = h5py.File(snapshot, 'r')
    rho  = f['/PartType0/Density'][:]   #(1e10 Msun/h)/(kpc/h)^3
    ne   = f['/PartType0/ElectronAbundance'][:] 
    SFR  = f['PartType0/StarFormationRate'][:]
    f.close()

    # formula is 0.76*ne*rho/m_proton
    # rho units are (Msun/h)/(kpc/h)^3;  1e19*2e30/(3.1e24)^3/3e-55
    factor = 1e10*Msun/kpc**3/m_proton

    indexes = np.where(SFR>0.0)
    n_e = factor * 0.76 * ne * rho #electrons*h^2/cm^3
    n_e[indexes] = 0.0 #put electron density to 0 for star-forming particles

    return n_e
########################################################################################

# This routine computes different properties of SO halos such as mass in each component
# the temperature of the halo, number of particles...etc.
# snapshot --------> snapshot
# f_subfind -------> SUBFIND catalague
# cell_size -------> particles/halos will be places in a grid. This is to speed up the
#                    calculation. The voxel size should be larger than the radius of the
#                    largest halo in the snapshot
# root_out --------> output folder
def SO_properties(snapshot, f_subfind, cell_size, root_out):

    # check if subfind catalogue exists
    if not(os.path.exists(f_subfind)):  return 0

    # read the header and get BoxSize, redshift and masses
    f        = h5py.File(snapshot, 'r')
    BoxSize  = f['Header'].attrs[u'BoxSize']/1e3    #Mpc/h
    redshift = f['Header'].attrs[u'Redshift']
    Masses   = f['Header'].attrs[u'MassTable']*1e10 #Msun/h
    f.close()

    # get the name of the output file
    fout = '%s/SO_z=%.2f.txt'%(root_out, redshift)
    if os.path.exists(fout):  return 0

    ############################## HALOS ###############################
    # read the halo catalogue
    f = h5py.File(f_subfind, 'r')
    halo_pos  = f['Group/GroupPos'][:]/1e3           #Mpc/h
    halo_mass = f['Group/Group_M_TopHat200'][:]*1e10 #Msun/h
    halo_R    = f['Group/Group_R_TopHat200'][:]/1e3  #Mpc/h
    f.close()
    
    # keep only halos with radii larger than 0
    indexes   = np.where(halo_R>0)[0]
    halo_pos  = halo_pos[indexes]
    halo_mass = halo_mass[indexes]
    halo_R    = halo_R[indexes]

    # sort halo positions (not really neccesary, but positions will be closer in memory)
    data = SL.sort_3D_pos(halo_pos, BoxSize, cell_size, return_indexes=True, 
                          return_offset=False)
    halo_pos  = data.pos_sorted
    halo_R    = halo_R[data.indexes]
    halo_mass = halo_mass[data.indexes]
    ######################################################################

    ############################# PARTICLES ##############################
    # open the snapshot
    f = h5py.File(snapshot, 'r')

    ##### black-holes #####
    pos_bh  = f['/PartType5/Coordinates'][:].astype(np.float32)/1e3 #Mpc/h
    mass_bh = f['/PartType5/Masses'][:]*1e10                        #Msun/h
    Nbh     = np.ones(pos_bh.shape[0], dtype=np.float32)

    # sort the positions
    data = SL.sort_3D_pos(pos_bh, BoxSize, cell_size, return_indexes=True, 
                          return_offset=True)
    pos_bh    = data.pos_sorted
    mass_bh   = mass_bh[data.indexes]
    offset_bh = data.offset;  del data

    # compute the gas mass and number of gas particles
    halo_Mbh = np.zeros(halo_R.shape[0], dtype=np.float64)
    halo_Nbh = np.zeros(halo_R.shape[0], dtype=np.float64)
    HIL.HI_mass_SO(halo_pos, halo_R, pos_bh, mass_bh, offset_bh, halo_Mbh, BoxSize)
    HIL.HI_mass_SO(halo_pos, halo_R, pos_bh, Nbh,     offset_bh, halo_Nbh, BoxSize)
    del pos_bh, mass_bh, Nbh, offset_bh
    ######################

    ##### stars #####
    pos_s  = f['/PartType4/Coordinates'][:].astype(np.float32)/1e3 #Mpc/h
    mass_s = f['/PartType4/Masses'][:]*1e10                        #Msun/h
    Ns     = np.ones(pos_s.shape[0], dtype=np.float32)

    # sort the positions
    data = SL.sort_3D_pos(pos_s, BoxSize, cell_size, return_indexes=True, 
                          return_offset=True)
    pos_s    = data.pos_sorted
    mass_s   = mass_s[data.indexes]
    offset_s = data.offset;  del data

    # compute the gas mass and number of gas particles
    halo_Ms = np.zeros(halo_R.shape[0], dtype=np.float64)
    halo_Ns = np.zeros(halo_R.shape[0], dtype=np.float64)
    HIL.HI_mass_SO(halo_pos, halo_R, pos_s, mass_s, offset_s, halo_Ms, BoxSize)
    HIL.HI_mass_SO(halo_pos, halo_R, pos_s, Ns,     offset_s, halo_Ns, BoxSize)
    del pos_s, mass_s, Ns, offset_s
    ###################

    ##### CDM #####
    pos_c  = f['/PartType1/Coordinates'][:].astype(np.float32)/1e3   #Mpc/h
    if '/PartType1/Masses' in f:  
        mass_c = f['/PartType1/Masses'][:]*1e10                      #Msun/h
    else: 
        mass_c = np.ones(pos_c.shape[0], dtype=np.float32)*Masses[1] #Msunh
    Nc     = np.ones(pos_c.shape[0], dtype=np.float32)

    # sort the positions
    data = SL.sort_3D_pos(pos_c, BoxSize, cell_size, return_indexes=True, 
                          return_offset=True)
    pos_c    = data.pos_sorted
    mass_c   = mass_c[data.indexes]
    offset_c = data.offset;  del data

    # compute the gas mass and number of gas particles
    halo_Mc = np.zeros(halo_R.shape[0], dtype=np.float64)
    halo_Nc = np.zeros(halo_R.shape[0], dtype=np.float64)
    HIL.HI_mass_SO(halo_pos, halo_R, pos_c, mass_c, offset_c, halo_Mc, BoxSize)
    HIL.HI_mass_SO(halo_pos, halo_R, pos_c, Nc,     offset_c, halo_Nc, BoxSize)
    del pos_c, mass_c, Nc, offset_c
    ###################

    ##### gas #####
    pos_g  = f['/PartType0/Coordinates'][:].astype(np.float32)/1e3 #Mpc/h
    mass_g = f['/PartType0/Masses'][:]*1e10                        #Msun/h
    SFR    = f['/PartType0/StarFormationRate'][:]
    T      = temperature(snapshot)
    Ng     = np.ones(pos_g.shape[0], dtype=np.float32)

    # sort the positions
    data = SL.sort_3D_pos(pos_g, BoxSize, cell_size, return_indexes=True, 
                          return_offset=True)
    pos_g    = data.pos_sorted
    mass_g   = mass_g[data.indexes]
    SFR      = SFR[data.indexes]
    T        = T[data.indexes]
    offset_g = data.offset;  del data

    # compute the gas mass and number of gas particles
    halo_Mg = np.zeros(halo_R.shape[0], dtype=np.float64)
    halo_Ng = np.zeros(halo_R.shape[0], dtype=np.float64)
    HIL.HI_mass_SO(halo_pos, halo_R, pos_g, mass_g, offset_g, halo_Mg, BoxSize)
    HIL.HI_mass_SO(halo_pos, halo_R, pos_g, Ng,     offset_g, halo_Ng, BoxSize)
    del offset_g, Ng
    ###############

    ##### non-star forming gas #####
    indexes   = np.where(SFR<=0.0)[0];  del SFR
    pos_nsfg  = pos_g[indexes];         del pos_g
    mass_nsfg = mass_g[indexes];        del mass_g
    T_nsfg    = T[indexes];             del T
    Nnsfg     = np.ones(pos_nsfg.shape[0], dtype=np.float32)

    # sort the positions
    data = SL.sort_3D_pos(pos_nsfg, BoxSize, cell_size, return_indexes=True, 
                          return_offset=True)
    pos_nsfg    = data.pos_sorted
    mass_nsfg   = mass_nsfg[data.indexes]
    T_nsfg      = T_nsfg[data.indexes]
    offset_nsfg = data.offset;  del data

    # compute the gas mass and number of gas particles
    halo_Mnsfg  = np.zeros(halo_R.shape[0], dtype=np.float64)
    halo_TMnsfg = np.zeros(halo_R.shape[0], dtype=np.float64)
    halo_Nnsfg  = np.zeros(halo_R.shape[0], dtype=np.float64)
    HIL.HI_mass_SO(halo_pos, halo_R, pos_nsfg, mass_nsfg,        offset_nsfg, 
                   halo_Mnsfg, BoxSize)
    HIL.HI_mass_SO(halo_pos, halo_R, pos_nsfg, mass_nsfg*T_nsfg, offset_nsfg, 
                   halo_TMnsfg, BoxSize)
    HIL.HI_mass_SO(halo_pos, halo_R, pos_nsfg, Nnsfg,            offset_nsfg, 
                   halo_Nnsfg, BoxSize)
    del pos_nsfg, mass_nsfg, Nnsfg, offset_nsfg
    ################################
    f.close()

    # check that masses from routine give similar results that SUBFIND
    ratio = (halo_Mg + halo_Mc + halo_Ms + halo_Mbh)/halo_mass
    print(np.min(ratio), np.max(ratio))
    
    # compute average halo gas temperature
    halo_Mnsfg[np.where(halo_Mnsfg==0.0)[0]] = 1.0
    halo_T = halo_TMnsfg/halo_Mnsfg
    halo_Mnsfg[np.where(halo_Mnsfg==1.0)[0]] = 0.0

    # save results to file
    np.savetxt(fout, np.transpose([halo_mass, halo_R, halo_pos[:,0], halo_pos[:,1], 
                                   halo_pos[:,2], halo_Mg, halo_Mc, halo_Ms, halo_Mbh,
                                   halo_T, halo_Ng, halo_Nc, halo_Ns, halo_Nbh, 
                                   halo_Nnsfg]))
########################################################################################

########################################################################################
# This routine computes the mean and std of different files
# root -----------> folder where data is
# sim ------------> simulation considered: 'SIMBA' or 'IllustrisTNG'
# name -----------> The generic name of the files
# realizations ---> array with the realizations to use for the mean and std
# first ----------> integer representing the first column to read in the files
# second ---------> integer representing the second column to read in the files
# fout -----------> name of output file
def mean_std(root, sim, name, realizations, first, second, fout, verbose=False):
    
    # read the first realization and get dimensions of data vector
    fin = '%s/%s/%s/%s'%(root, sim, realizations[0], name)
    data_real = np.loadtxt(fin);  X,Y = data_real[:,first], data_real[:,second]
    data = np.zeros((len(realizations), len(X)), dtype=np.float64)

    # read all files and fill up the array
    count = 0
    for i in realizations:
        fin = '%s/%s/%s/%s'%(root, sim, i, name)
        if not(os.path.exists(fin)):  continue
        data_real = np.loadtxt(fin);  data[count] = data_real[:,second]
        count += 1
    if verbose:  print('Found %d realizations for %s'%(count,sim))
    data = data[:count]
    
    # save mean and std to file
    np.savetxt(fout, np.transpose([X, np.mean(data, axis=0), np.std(data, axis=0),
                                   np.percentile(data, 84, axis=0), 
                                   np.percentile(data, 16, axis=0),
                                   np.median(data, axis=0)]))
########################################################################################

########################################################################################
# This routine computes the mean and std of the halos temperature
# root -----------> folder where data is
# sim ------------> simulation considered: 'SIMBA' or 'IllustrisTNG'
# name -----------> The generic name of the files
# realizations ---> array with the realizations to use for the mean and std
# Mmin_T ---------> minimum halo mass
# Mmax_T ---------> maximum halo mass
# bins_T ---------> number of bins between Mmin_T and Mmax_T
# fout -----------> name of output file
def mean_std_T(root, sim, name, realizations, Mmin_T, Mmax_T, bins_T, 
               fout, verbose=False):

    # define the arrays containing the temperatures and number of halos
    bins_mass = np.logspace(np.log10(Mmin_T), np.log10(Mmax_T), bins_T+1)
    mass_mean = 0.5*(bins_mass[1:] + bins_mass[:-1])
    log_min, log_max = np.log10(Mmin_T), np.log10(Mmax_T)

    # define the structure containing the halo temperatures
    T = []
    for i in range(bins_T):  T.append([])

    # do a loop over all realizations
    count = 0
    for i in realizations:

        # read data
        fin = '%s/%s/%s/%s'%(root, sim, i, name)
        if not(os.path.exists(fin)):  continue
        data = np.loadtxt(fin)  
        halo_M, halo_T, Nnsfg = data[:,0], data[:,9], data[:,14]
            
        # only consider halos with more than 50 gas particles
        indexes = np.where(Nnsfg>50.0)[0]
        halo_M  = halo_M[indexes]
        halo_T  = halo_T[indexes]
        log_M   = np.log10(halo_M)

        # do a loop over the different halos
        for j in range(len(halo_M)):
            index = (log_M[j] - log_min)/(log_max - log_min)*bins_T
            if index<0.0 or index>bins_T:  continue
            index = int(index)
            T[index].append(halo_T[j])

        count +=1
    print('Found %d realization for %s'%(count,sim))

    # save results to file
    f = open(fout, 'w')
    for i in range(bins_T):
        array = np.array(T[i])
        f.write('%.5e %.5e %.5e %.5e %.5e %.5e\n'%(mass_mean[i], np.mean(array),
                                            np.std(array), np.percentile(array,84),
                                            np.percentile(array,16), np.median(array)))
    f.close()
########################################################################################

########################################################################################
# This routine computes the mean and std of galaxy:
# 1) radii
# 2) black hole masses
# 3) SFR
# 4) maximum velocities
# root -----------> folder where data is
# sim ------------> simulation considered: 'SIMBA' or 'IllustrisTNG'
# name -----------> The generic name of the files
# realizations ---> array with the realizations to use for the mean and std
# SMmin ----------> minimum stellar mass
# SMmax ----------> maximum stellar mass
# bins -----------> number of bins between SMmin and SMmax
# fout -----------> name of output file
def mean_std_SM(root, sim, name, realizations, SMmin, SMmax, bins,
                fout_R, fout_BH, fout_SFR, fout_V, verbose=False):

    # define the arrays containing the temperatures and number of halos
    bins_SM = np.logspace(np.log10(SMmin), np.log10(SMmax), bins+1)
    SM_mean = 0.5*(bins_SM[1:] + bins_SM[:-1])

    # define the lists containing the galaxy properties
    R, BH, SFR, V = [], [], [], []
    for i in range(bins):
        R.append([]);  BH.append([]);  SFR.append([]);  V.append([])

    # do a loop over all realizations
    count = 0
    for i in realizations:

        # read data
        fin = '%s/%s/%s/%s'%(root, sim, i, name)
        if not(os.path.exists(fin)):  continue
        f    = h5py.File(fin, 'r')
        SM   = f['Subhalo/SubhaloMassType'][:][:,4]*1e10   #Msun/h
        Rp   = f['Subhalo/SubhaloHalfmassRadType'][:][:,4] #kpc/h
        BHp  = f['Subhalo/SubhaloBHMass'][:]*1e10          #Msun/h
        SFRp = f['Subhalo/SubhaloSFR'][:]
        Vp   = f['Subhalo/SubhaloVmax'][:]                 #km/s
        f.close()
        
        # consider only galaxies within given stellar mass range
        indexes = np.where((SM>=SMmin) & (SM<SMmax))
        SM   = SM[indexes]
        Rp   = Rp[indexes]
        BHp  = BHp[indexes]
        SFRp = SFRp[indexes]
        Vp   = Vp[indexes]

        # do a loop over all galaxies
        index = (np.log10(SM)-np.log10(SMmin))/(np.log10(SMmax)-np.log10(SMmin))*bins
        index = np.int32(index)
        for j in range(len(SM)):
            R[index[j]].append(Rp[j])
            BH[index[j]].append(BHp[j])
            SFR[index[j]].append(SFRp[j])
            V[index[j]].append(Vp[j])

        count += 1
    print('found %d realization for %s'%(count,sim))

    f1 = open(fout_R,   'w');  f2 = open(fout_BH, 'w');  
    f3 = open(fout_SFR, 'w');  f4 = open(fout_V,  'w')
    for i in range(bins):
        array = np.array(R[i])
        f1.write('%.5e %.5e %.5e %.5e %.5e %.5e\n'%(SM_mean[i], np.mean(array),
                                        np.std(array), np.percentile(array,84),
                                        np.percentile(array,16), np.median(array)))
        array = np.array(BH[i])
        f2.write('%.5e %.5e %.5e %.5e %.5e %.5e\n'%(SM_mean[i], np.mean(array),
                                        np.std(array), np.percentile(array,84),
                                        np.percentile(array,16), np.median(array)))
        array = np.array(SFR[i])
        indexes = np.where(array>0.0)[0]  #select galaxies with SFR>0
        array2 = array[indexes]
        f3.write('%.5e %.5e %.5e %.5e %.5e %.5e %.5e %.5e %.5e %.5e %.5e\n'%(SM_mean[i], np.mean(array),
                                        np.std(array), np.percentile(array,84),
                                        np.percentile(array,16), np.median(array),
                                        np.mean(array2), np.std(array2), 
                                        np.percentile(array2,84), 
                                        np.percentile(array2,16), np.median(array2)))
        array = np.array(V[i])
        f4.write('%.5e %.5e %.5e %.5e %.5e %.5e\n'%(SM_mean[i], np.mean(array),
                                        np.std(array), np.percentile(array,84),
                                        np.percentile(array,16), np.median(array)))
    f1.close();  f2.close();  f3.close();  f4.close()

########################################################################################

########################################################################################
# This routine computes the mean and std of galaxy:
# 1) radii
# 2) black hole masses
# 3) SFR
# 4) maximum velocities
# root -----------> folder where data is
# sim ------------> simulation considered: 'SIMBA' or 'IllustrisTNG'
# name -----------> The generic name of the files
# realizations ---> array with the realizations to use for the mean and std
# SMmin ----------> minimum stellar mass
# SMmax ----------> maximum stellar mass
# bins -----------> number of bins between SMmin and SMmax
# fout -----------> name of output file
def mean_std_SM_old(root, sim, name, realizations, SMmin, SMmax, bins,
                    fout_R, fout_BH, fout_SFR, fout_V, verbose=False):
    
    #if os.path.exists(fout_R) and os.path.exists(fout_BH) and\
    #   os.path.exists(fout_Z) and os.path.exists(fout_V):  return 0

    # define the arrays containing the temperatures and number of halos
    bins_SM = np.logspace(np.log10(SMmin), np.log10(SMmax), bins+1)
    SM_mean = 0.5*(bins_SM[1:] + bins_SM[:-1])
    R    = np.zeros(bins, dtype=np.float64)
    R2   = np.zeros(bins, dtype=np.float64)
    BH   = np.zeros(bins, dtype=np.float64)
    BH2  = np.zeros(bins, dtype=np.float64)
    SFR  = np.zeros(bins, dtype=np.float64)
    SFR2 = np.zeros(bins, dtype=np.float64)
    V    = np.zeros(bins, dtype=np.float64)
    V2   = np.zeros(bins, dtype=np.float64)
    N    = np.zeros(bins, dtype=np.float64)
    N_BH = np.zeros(bins, dtype=np.float64)

    # do a loop over all realizations
    count = 0
    for i in realizations:

        # read data
        fin = '%s/%s/%d/%s'%(root, sim, i, name)
        if not(os.path.exists(fin)):  continue
        f    = h5py.File(fin, 'r')
        SM   = f['Subhalo/SubhaloMassType'][:][:,4]*1e10   #Msun/h
        Rp   = f['Subhalo/SubhaloHalfmassRadType'][:][:,4] #kpc/h
        BHp  = f['Subhalo/SubhaloBHMass'][:]*1e10          #Msun/h
        SFRp = f['Subhalo/SubhaloSFR'][:]
        Vp   = f['Subhalo/SubhaloVmax'][:]                 #km/s
        f.close()

        ids = np.where(SM>3e11)[0]
        if len(ids)>0:
            f = open('BH_test.txt', 'a')
            for i in range(len(ids)):
                f.write('%.4e\n'%BHp[ids][i])
            f.close()


        # radii
        R  += np.histogram(SM, bins_SM, weights=Rp)[0]
        R2 += np.histogram(SM, bins_SM, weights=Rp**2)[0]

        # black-holes masses
        indexes = np.where(BHp>1e8)[0]
        BH   += np.histogram(SM[indexes], bins_SM, weights=BHp[indexes])[0]
        BH2  += np.histogram(SM[indexes], bins_SM, weights=BHp[indexes]**2)[0]
        N_BH += np.histogram(SM[indexes], bins_SM)[0]
        #BH  += np.histogram(SM, bins_SM, weights=BHp)[0]
        #BH2 += np.histogram(SM, bins_SM, weights=BHp**2)[0]

        # SFR
        SFR  += np.histogram(SM, bins_SM, weights=SFRp)[0]
        SFR2 += np.histogram(SM, bins_SM, weights=SFRp**2)[0]

        # maximum velocity
        indexes = np.where(Vp!=np.inf)[0]
        V  += np.histogram(SM[indexes], bins_SM, weights=Vp[indexes])[0]
        V2 += np.histogram(SM[indexes], bins_SM, weights=Vp[indexes]**2)[0]

        # number of galaxies
        N  += np.histogram(SM, bins_SM)[0]
        count += 1
    print('found %d realization for %s'%(count,sim))

    # find average and standard deviation; save results to file
    R   = R/N;    dR   = np.sqrt(R2/N -   R**2)
    BH  = BH/N_BH;   dBH  = np.sqrt(BH2/N_BH -  BH**2)
    #BH  = BH/N;   dBH  = np.sqrt(BH2/N -  BH**2)
    SFR = SFR/N;  dSFR = np.sqrt(SFR2/N - SFR**2)
    V   = V/N;    dV   = np.sqrt(V2/N -   V**2)

    np.savetxt(fout_R,   np.transpose([SM_mean, R,   dR,   N]))
    np.savetxt(fout_BH,  np.transpose([SM_mean, BH,  dBH,  N]))
    np.savetxt(fout_SFR, np.transpose([SM_mean, SFR, dSFR, N]))
    np.savetxt(fout_V,   np.transpose([SM_mean, V,   dV,   N]))
########################################################################################

########################################################################################
# This routine computes the mean and std of the halos temperature
# root -----------> folder where data is
# sim ------------> simulation considered: 'SIMBA' or 'IllustrisTNG'
# name -----------> The generic name of the files
# realizations ---> array with the realizations to use for the mean and std
# Mmin_T ---------> minimum halo mass
# Mmax_T ---------> maximum halo mass
# bins_T ---------> number of bins between Mmin_T and Mmax_T
# fout -----------> name of output file
def mean_std_T_old(root, sim, name, realizations, Mmin_T, Mmax_T, bins_T, 
                   fout, verbose=False):

    # define the arrays containing the temperatures and number of halos
    bins_mass = np.logspace(np.log10(Mmin_T), np.log10(Mmax_T), bins_T+1)
    mass_mean = 0.5*(bins_mass[1:] + bins_mass[:-1])
    T  = np.zeros(bins_T, dtype=np.float64)
    T2 = np.zeros(bins_T, dtype=np.float64)
    N  = np.zeros(bins_T, dtype=np.float64)

    # do a loop over all realizations
    count = 0
    for i in realizations:

        # read data
        fin = '%s/%s/%d/%s'%(root, sim, i, name)
        if not(os.path.exists(fin)):  continue
        data = np.loadtxt(fin)  
        halo_M, halo_T, Nnsfg = data[:,0], data[:,9], data[:,14]
        #if i in [243,408,424] and sim=='SIMBA':
        #    continue
            #if np.any(halo_T>1e9):  
            #print(sim,i)
            #continue
            
        # only consider halos with more than 50 gas particles
        indexes = np.where(Nnsfg>50.0)[0]
        halo_M  = halo_M[indexes]
        halo_T  = halo_T[indexes]
        
        T  += np.histogram(halo_M, bins_mass, weights=halo_T)[0]
        T2 += np.histogram(halo_M, bins_mass, weights=halo_T**2)[0]
        N  += np.histogram(halo_M, bins_mass)[0]
        count += 1
    print('found %d realization for %s'%(count,sim))

    # find average and standard deviation;  save results to file
    T  = T/N
    dT = np.sqrt(T2/N - T**2)
    np.savetxt(fout, np.transpose([mass_mean, T, dT, N]))
########################################################################################

########################################################################################
# This routine will read data from a snapshot and will return the relevant quantities
# for a given field. For instance, for the field T, it will return the positions, 
# masses, and temperature of gas particles.
# snapshot -------> the name of considered snapshot
# field ----------> the considered field, e.g. 'T', or 'Mgas'
# sim ------------> 'IllustrisTNG' or 'SIMBA'
def field_properties(snapshot, field, sim):

    # check whether the snapshot exists
    if not(os.path.exists(snapshot)):  raise Exception('%s does not exists'%snapshot)

    # check that the field is included
    if field not in ['Mgas', 'Mcdm', 'Mtot', 'Mstar', 'T', 'Z', 'P', 
                     'ne', 'HI', 'Vgas', 'Vcdm', 'B', 'MgFe']:
        raise Exception('%s not implemented'%field)

    # read the main properties of the snapshot
    f = h5py.File(snapshot, 'r')
    BoxSize  = f['Header'].attrs[u'BoxSize']/1e3 #Mpc/h
    redshift = f['Header'].attrs[u'Redshift']
    h        = f['Header'].attrs[u'HubbleParam']
    Masses   = f['Header'].attrs[u'MassTable'][:]*1e10 #Msun/h
    Ntot     = f['Header'].attrs[u'NumPart_ThisFile']

    # read the positions and masses of the gas particles for all these fields
    if field in ['Mgas', 'T', 'Z', 'P', 'ne', 'HI', 'Vgas', 'B', 'MgFe', 'Mtot']:
        pos_g = f['PartType0/Coordinates'][:]/1e3 #Mpc/h
        pos_g = pos_g.astype(np.float32)          #Mpc/h
        Mg    = f['PartType0/Masses'][:]*1e10     #Msun/h  

    # read the positions and masses of the dark matter particles for these fields
    if field in ['Mcdm', 'Vcdm', 'Mtot']:
        pos_c = f['PartType1/Coordinates'][:]/1e3 #Mpc/h
        pos_c = pos_c.astype(np.float32)          #Mpc/h
        if sim=='IllustrisTNG':
            Mc = np.ones(pos_c.shape[0], dtype=np.float32)*Masses[1] #Msun/h
        if sim=='SIMBA':
            Mc = f['PartType1/Masses'][:]*1e10 #Msun/h

    # read the positions and masses of the star particles for these fields
    if field in ['Mstar', 'Mtot']:
        if sim=='IllustrisTNG':
            pos_sw = (f['PartType4/Coordinates'][:]/1e3).astype(np.float32)
            Msw    = f['PartType4/Masses'][:]*1e10     #Msun/h
            #Age    = f['PartType4/GFM_StellarFormationTime'][:] #stars have Age>0
            #indexes = np.where(Age>0)[0]
            #pos_s  = pos_sw[indexes] # FIXME unused! -- Paco double check
            #Ms     = Msw[indexes]
        else:
            pos_sw = (f['PartType4/Coordinates'][:]/1e3).astype(np.float32) #Mpc/h
            Msw    = f['PartType4/Masses'][:]*1e10 #Msun/h
            #pos_s  = np.copy(pos_sw)
            #Ms     = np.copy(Msw)

    # gas mass
    if field=='Mgas':  
        return pos_g, Mg

    # dark matter mass
    if field=='Mcdm':
        return pos_c, Mc

    # stellar mass
    if field=='Mstar':
        return pos_sw, Msw

    # gas temperature
    if field=='T':
        T  = temperature(snapshot) #K
        return pos_g, Mg, T

    # gas metallicity
    if field=='Z':
        if sim=='IllustrisTNG':
            Z = f['PartType0/GFM_Metallicity'][:] #metallicity
        if sim=='SIMBA':
            Z = f['PartType0/Metallicity'][:,0]+8e-10 #metallicity
        return pos_g, Mg, Z

    # gas pressure
    if field=='P':
        P = pressure(snapshot) #Msun*(km/s)^2/kpc^3
        return pos_g, Mg, P

    # electron density
    if field=='ne':
        rho_g = f['PartType0/Density'][:]*1e19    #(Msun/h)/(Mpc/h)^3
        Vol_g = Mg/rho_g;  del rho_g              #(Mpc/h)^3
        ne = electron_density(snapshot)*Vol_g #electrons/h
        return pos_g, Mg, ne

    # neutral hydrogen
    if field=='HI':
        HI = HI_mass(snapshot, TREECOOL_file, sim)
        return pos_g, HI

    # gas velocity
    if field=='Vgas':
        vel_g = f['PartType0/Velocities'][:]/np.sqrt(1.0+redshift)
        vel_g = np.sqrt(vel_g[:,0]**2 + vel_g[:,1]**2 + vel_g[:,2]**2)
        return pos_g, Mg, vel_g

    # dark matter velocity
    if field=='Vcdm':
        vel_c = f['PartType1/Velocities'][:]/np.sqrt(1.0+redshift)
        vel_c = np.sqrt(vel_c[:,0]**2 + vel_c[:,1]**2 + vel_c[:,2]**2)
        return pos_c, Mc, vel_c

    # magnetic fields
    if field=='B' and sim=='IllustrisTNG':
        B = np.linalg.norm(f['PartType0/MagneticField'][:]*2.6e-6*h/(1.0+redshift)**2, 
                           axis=-1) #Gauss
        return pos_g, Mg, B

    # magnesium over iron
    if field=='MgFe':
        if sim=='IllustrisTNG':
            Mmg = f['PartType0/GFM_Metals'][:,6]*Mg #magnesium
            Mfe = f['PartType0/GFM_Metals'][:,8]*Mg #iron
        if sim=='SIMBA':
            Mmg = (f['PartType0/Metallicity'][:,6]+1e-10)*Mg  #magnesium
            Mfe = (f['PartType0/Metallicity'][:,10]+1e-10)*Mg #iron
        return pos_g, Mmg, Mfe

    # total matter mass
    if field=='Mtot':
        if Ntot[5]>0:  
            pos_bh = f['PartType5/Coordinates'][:]/1e3 #Mpc/h
            pos_bh = pos_bh.astype(np.float32)         #Mpc/h
            Mbh    = f['PartType5/Masses'][:]*1e10     #Msun/h        
            return pos_g, Mg, pos_c, Mc, pos_sw, Msw, pos_bh, Mbh
        else:
            return pos_g, Mg, pos_c, Mc, pos_sw, Msw, pos_bh, Mbh

    # close file
    f.close()



import torch
import torch.nn as nn

def weights_init(m):
    """custom weights initialization
    """
    classname = m.__class__.__name__
    if classname.find('Conv') != -1:
        m.weight.data.normal_(0.0, 0.02)
    elif classname.find('BatchNorm') != -1:
        m.weight.data.normal_(1.0, 0.02)
        m.bias.data.fill_(0)

# ConvTranspose2d(channels_in, channels_out, kernel, stride, padding)
class Generator_64(nn.Module):
    def __init__(self, Z_DIM, G_HIDDEN):
        super(Generator_64, self).__init__()
        self.main = nn.Sequential(
            # 1st layer (input: 100x1x1 ----> output: 512x4x4)
            nn.ConvTranspose2d(Z_DIM, G_HIDDEN * 8, 4, 1, 0, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 8),
            nn.ReLU(True),
            # 2nd layer (input: 512x4x4 ----> output: 256x8x8)
            nn.ConvTranspose2d(G_HIDDEN * 8, G_HIDDEN * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 4),
            nn.ReLU(True),
            # 3rd layer (input: 256x8x8 ----> output: 128x16x16)
            nn.ConvTranspose2d(G_HIDDEN * 4, G_HIDDEN * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 2),
            nn.ReLU(True),
            # 4th layer (input: 128x16x16 ----> output: 64x32x32)
            nn.ConvTranspose2d(G_HIDDEN * 2, G_HIDDEN, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN),
            nn.ReLU(True),
            # output layer (input: 64x32x32 ----> 1x64x64)
            nn.ConvTranspose2d(G_HIDDEN, 1, 4, 2, 1, bias=False),
            nn.Tanh()
        )

    def forward(self, input):
        return self.main(input)


#Conv2d(channels_in, channels_out, kernel, stride, padding)
class Discriminator_64(nn.Module):
    def __init__(self, D_HIDDEN):
        super(Discriminator_64, self).__init__()
        self.main = nn.Sequential(
            # 1st layer (input: 1x64x64 ----> output: 64x32x32)
            nn.Conv2d(1, D_HIDDEN, 4, 2, 1, bias=False),
            nn.LeakyReLU(0.2, inplace=True),
            # 2nd layer (input: 64x32x32 ----> output: 128x16x16)
            nn.Conv2d(D_HIDDEN, D_HIDDEN * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 2),
            nn.LeakyReLU(0.2, inplace=True),
            # 3rd layer (input: 128x16x16 ----> output: 256x8x8)
            nn.Conv2d(D_HIDDEN * 2, D_HIDDEN * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 4),
            nn.LeakyReLU(0.2, inplace=True),
            # 4th layer (input: 256x8x8 ----> output: 512x4x4)
            nn.Conv2d(D_HIDDEN * 4, D_HIDDEN * 8, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 8),
            nn.LeakyReLU(0.2, inplace=True),
            # output layer (input: 512x4x4 ----> output: 1x1x1)
            nn.Conv2d(D_HIDDEN * 8, 1, 4, 1, 0, bias=False),
            nn.Sigmoid() #Returns a number from 0 to 1: probability
        )

    def forward(self, input):
        return self.main(input).view(-1, 1).squeeze(1)







# ConvTranspose2d(channels_in, channels_out, kernel, stride, padding)
class Generator_128(nn.Module):
    def __init__(self, Z_DIM, G_HIDDEN):
        super(Generator_128, self).__init__()
        self.main = nn.Sequential(
            # 1st layer (input: 100x1x1 ----> output: 512x8x8)
            nn.ConvTranspose2d(Z_DIM, G_HIDDEN * 8, 8, 1, 0, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 8),
            nn.ReLU(True),
            # 2nd layer (input: 512x8x8 ----> output: 256x16x16)
            nn.ConvTranspose2d(G_HIDDEN * 8, G_HIDDEN * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 4),
            nn.ReLU(True),
            # 3rd layer (input: 256x16x16 ----> output: 128x32x32)
            nn.ConvTranspose2d(G_HIDDEN * 4, G_HIDDEN * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN * 2),
            nn.ReLU(True),
            # 4th layer (input: 128x32x32 ----> output: 64x64x64)
            nn.ConvTranspose2d(G_HIDDEN * 2, G_HIDDEN, 4, 2, 1, bias=False),
            nn.BatchNorm2d(G_HIDDEN),
            nn.ReLU(True),
            # output layer (input: 64x64x64 ----> 1x128x128)
            nn.ConvTranspose2d(G_HIDDEN, 1, 4, 2, 1, bias=False),
        )

    def forward(self, input):
        return self.main(input)


#Conv2d(channels_in, channels_out, kernel, stride, padding)
class Discriminator_128(nn.Module):
    def __init__(self, D_HIDDEN):
        super(Discriminator_128, self).__init__()
        self.main = nn.Sequential(
            # 1st layer (input: 1x128x128 ----> output: 64x32x32)
            nn.Conv2d(1, D_HIDDEN, 4, 4, 0, bias=False),
            nn.LeakyReLU(0.2, inplace=True),
            # 2nd layer (input: 64x32x32 ----> output: 128x16x16)
            nn.Conv2d(D_HIDDEN, D_HIDDEN * 2, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 2),
            nn.LeakyReLU(0.2, inplace=True),
            # 3rd layer (input: 128x16x16 ----> output: 256x8x8)
            nn.Conv2d(D_HIDDEN * 2, D_HIDDEN * 4, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 4),
            nn.LeakyReLU(0.2, inplace=True),
            # 4th layer (input: 256x8x8 ----> output: 512x4x4)
            nn.Conv2d(D_HIDDEN * 4, D_HIDDEN * 8, 4, 2, 1, bias=False),
            nn.BatchNorm2d(D_HIDDEN * 8),
            nn.LeakyReLU(0.2, inplace=True),
            # output layer (input: 512x4x4 ----> output: 1x1x1)
            nn.Conv2d(D_HIDDEN * 8, 1, 4, 1, 0, bias=False),
            nn.Sigmoid() #Returns a number from 0 to 1: probability
        )

    def forward(self, input):
        return self.main(input).view(-1, 1).squeeze(1)




from pylab import *
import numpy as np
from matplotlib.ticker import ScalarFormatter
import matplotlib.gridspec as gridspec
from mpl_toolkits.axes_grid1.inset_locator import mark_inset
from matplotlib.ticker import AutoMinorLocator
from matplotlib.colors import LogNorm
from matplotlib.patches import Ellipse
rcParams["mathtext.fontset"]='cm'

############################### figure ###########################
#fig=figure(figsize=(15,10))     #give dimensions to the figure
##################################################################

################################ INPUT #######################################
#axes range
##############################################################################

############################ subplots ############################
#gs = gridspec.GridSpec(2,1,height_ratios=[5,2])
#ax1=plt.subplot(gs[0])
#ax2=plt.subplot(gs[1])

#make a subplot at a given position and with some given dimensions
#ax2=axes([0.4,0.55,0.25,0.1])

#gs.update(hspace=0.0,wspace=0.4,bottom=0.6,top=1.05)
#subplots_adjust(left=None, bottom=None, right=None, top=None,
#                wspace=0.5, hspace=0.5)

#set minor ticks
#ax1.xaxis.set_minor_locator(AutoMinorLocator(4))
#ax1.yaxis.set_minor_locator(AutoMinorLocator(4))


#ax1.xaxis.set_major_formatter( NullFormatter() )   #unset x label 
#ax1.yaxis.set_major_formatter( NullFormatter() )   #unset y label

#ax1.set_xticks([])
#ax1.set_yticks([])


#ax1.get_yaxis().set_label_coords(-0.2,0.5)  #align y-axis for multiple plots
##################################################################

##################### special behaviour stuff ####################
#to show error missing error bars in log scale
#ax1.set_yscale('log',nonposy='clip')  #set log scale for the y-axis

#set the x-axis in %f format instead of %e
#ax1.xaxis.set_major_formatter(ScalarFormatter()) 

#set size of ticks
#ax1.tick_params(axis='both', which='major', labelsize=10)
#ax1.tick_params(axis='both', which='minor', labelsize=8)

#set the position of the ylabel 
#ax1.yaxis.set_label_coords(-0.2, 0.4)

#set yticks in scientific notation
#ax1.ticklabel_format(axis='y',style='sci',scilimits=(1,4))

#set the x-axis in %f format instead of %e
#formatter = matplotlib.ticker.FormatStrFormatter('$%.2e$') 
#ax1.yaxis.set_major_formatter(formatter) 

#add two legends in the same plot
#ax5 = ax1.twinx()
#ax5.yaxis.set_major_formatter( NullFormatter() )   #unset y label 
#ax5.legend([p1,p2],['0.0 eV','0.3 eV'],loc=3,prop={'size':14},ncol=1)

#set points to show in the yaxis
#ax1.set_yticks([0,1,2])

#highlight a zoomed region
#mark_inset(ax1, ax2, loc1=2, loc2=4, fc="none",edgecolor='purple')
##################################################################

############################ plot type ###########################
#standard plot
#p1,=ax1.plot(x,y,linestyle='-',marker='None')

#error bar plot with the minimum and maximum values of the error bar interval
#p1=ax1.errorbar(r,xi,yerr=[delta_xi_min,delta_xi_max],lw=1,fmt='o',ms=2,
#               elinewidth=1,capsize=5,linestyle='-') 

#filled area
#p1=ax1.fill_between([x_min,x_max],[1.02,1.02],[0.98,0.98],color='k',alpha=0.2)

#hatch area
#ax1.fill([x_min,x_min,x_max,x_max],[y_min,3.0,3.0,y_min],#color='k',
#         hatch='X',fill=False,alpha=0.5)

#scatter plot
#p1=ax1.scatter(k1,Pk1,c='b',edgecolor='none',s=8,marker='*')

#plot with markers
#pl4,=ax1.plot(ke3,Pk3/Pke3,marker='.',markevery=2,c='r',linestyle='None')

#set size of dashed lines
#ax.plot([0, 1], [0, 1], linestyle='--', dashes=(5, 1)) #length of 5, space of 1

#image plot
#cax = ax1.imshow(densities,cmap=get_cmap('jet'),origin='lower',
#           extent=[x_min, x_max, y_min, y_max],
#           #vmin=min_density,vmax=max_density)
#           norm = LogNorm(vmin=min_density,vmax=max_density))
#cbar = fig.colorbar(cax, ax2, ax=ax1, ticks=[-1, 0, 1]) #in ax2 colorbar of ax1
#cbar.set_label(r"$M_{\rm CSF}\/[h^{-1}M_\odot]$",fontsize=14,labelpad=-50)
#cbar.ax.tick_params(labelsize=10)  #to change size of ticks

#make a polygon
#polygon = Rectangle((0.4,50.0), 20.0, 20.0, edgecolor='purple',lw=0.5,
#                    fill=False)
#ax1.add_artist(polygon)
####################################################################

root = '/mnt/ceph/users/camels'

x_min, x_max = -0.1, 7.1
y_min = [4e-4, 4e-4, 4e-4, 4e-4, 4e-4, 4e-4]
y_max = [0.13, 0.13, 0.13, 0.13, 0.13, 0.13]



fig = figure(figsize=(15,8)) 
ax1 = fig.add_subplot(231)
ax2 = fig.add_subplot(234)
ax3 = fig.add_subplot(232)
ax4 = fig.add_subplot(235)
ax5 = fig.add_subplot(233)
ax6 = fig.add_subplot(236)

subplots_adjust(left=None, bottom=None, right=None, top=None,
                wspace=0.2, hspace=0.15)

fin   = '%s/Results/neural_nets/params_2_SFRH/SFRH_ML_pred.txt'%root
f_out = 'SFRH_NN_1parameter.pdf'

# read data
data = np.loadtxt(fin)
z = data[:,0]


for i,ax in enumerate([ax1,ax2,ax3,ax4,ax5,ax6]):
    #ax.set_xscale('log')
    ax.set_yscale('log')
    ax.set_xlim([x_min,x_max])
    ax.set_ylim([y_min[i],y_max[i]])

for ax in [ax2,ax4,ax6]:
    ax.set_xlabel(r'$z$',fontsize=18)

for ax in [ax1,ax2]:
    ax.set_ylabel(r'${\rm SFRH}\,[M_\odot{\rm yr^{-1}}{\rm Mpc}^{-3}]$',fontsize=18)

    
for start,end,ax in zip([1,  12, 23, 34, 45, 56],
                        [11, 22, 33, 44, 55, 66],
                        [ax1,ax2,ax3,ax4,ax5,ax6]):

    # read the SFRH
    for i,c in zip(np.arange(start,end,2),['red','orange','k','green','blue']):
        
        ax.plot(z,data[:,i+1],linestyle='-',marker='None',c=c)


#place a label in the plot
ax1.text(0.07, 0.07, r'$\Omega_{\rm m}$', fontsize=22, color='k', transform=ax1.transAxes)
ax2.text(0.07, 0.07, r'$\sigma_8$', fontsize=22, color='k', transform=ax2.transAxes)
ax3.text(0.07, 0.07, r'$A_1$', fontsize=22, color='k', transform=ax3.transAxes)
ax4.text(0.07, 0.07, r'$A_2$', fontsize=22, color='k', transform=ax4.transAxes)
ax5.text(0.07, 0.07, r'$A_3$', fontsize=22, color='k', transform=ax5.transAxes)
ax6.text(0.07, 0.07, r'$A_4$', fontsize=22, color='k', transform=ax6.transAxes)


savefig(f_out, bbox_inches='tight')
close(fig)












import numpy as np
import torch
import sys,os
sys.path.append('../')
import data as data
import architecture

#################################### INPUT ##########################################
root_in  = '/mnt/ceph/users/camels'
root_out = '/mnt/ceph/users/camels/Results/neural_nets/params_2_SFRH'
sim    = 'IllustrisTNG'
bins_SFRH = 100
realizations = 1000

# architecture parameters
h1 = 500
h2 = 2000
h3 = 2000
h4 = 100
wd = 1e-8
dr = 0.2

fmodel = '%s/model.pt'%root_out
fout   = '%s/SFRH_ML_pred.txt'%root_out
#####################################################################################

# get GPU if possible
if torch.cuda.is_available():
    print("CUDA Available")
    device = torch.device('cuda')
else:
    print("CUDA Not Available")
    device = torch.device('cpu')

# get the test dataset
params_sorted = np.loadtxt('%s/params_%s.txt'%(root_out,sim))
min_params    = np.min(params_sorted, axis=0)
max_params    = np.max(params_sorted, axis=0)

# generate the parameters                            
params = np.zeros((66,6), dtype=np.float64)

# set fiducial values
params[:,0] = 0.3
params[:,1] = 0.8
params[:,2] = 1.0
params[:,3] = 1.0
params[:,4] = 1.0
params[:,5] = 1.0

# set value of sims with different Omega_m                             
params[:11,  0] = np.linspace(0.1, 0.5, 11)
params[11:22,1] = np.linspace(0.6, 1.0, 11)
params[22:33,2] = np.logspace(np.log10(0.25), np.log10(4.00), 11)
params[33:44,3] = np.logspace(np.log10(0.25), np.log10(4.00), 11)
params[44:55,4] = np.logspace(np.log10(0.50), np.log10(2.00), 11)
params[55:66,5] = np.logspace(np.log10(0.50), np.log10(2.00), 11)
params = (params - min_params)/(max_params - min_params)
params = params.astype(np.float32)
params = torch.from_numpy(params)

# get the parameters of the trained model
model = architecture.model_1hl(6, h1, bins_SFRH, dr)
#model = architecture.model_3hl(6, h1, h2, h3, bins_SFRH, dr)
model.load_state_dict(torch.load(fmodel))
model.to(device=device)

# get prediction
offset = 0
model.eval()
with torch.no_grad():
    params = params.to(device)
    SFRH   = model(params)

# load all SFRH and compute mean and std
SFRH_all = data.read_all_SFRH(root_in, sim, realizations, bins_SFRH, root_out)
SFRH_all = np.log10(SFRH_all)
mean = np.mean(SFRH_all, axis=0, dtype=np.float64)
std  = np.std(SFRH_all,  axis=0, dtype=np.float64)

# de-normalize
SFRH = SFRH.cpu().numpy()
SFRH = 10**(SFRH*std + mean)

# save results to file
z = np.linspace(0.0, 7.0, bins_SFRH)
results = np.zeros((bins_SFRH, 67), dtype=np.float32)
results[:,0] = z
for i in range(1,67):
    results[:,i] = SFRH[i-1]
np.savetxt(fout, results)



import torch 
from torch.utils.data.dataset import Dataset
from torch.utils.data import DataLoader
import numpy as np
import sys, os, time


# This class creates the dataset 
class make_dataset():

    def __init__(self, mode, seed, f_images, grid, minimum=None, maximum=None, 
                 verbose=False):

        # read the data
        data = np.load(f_images) #[number of maps, height, width]

        # normalize maps
        data = np.log10(data)
        if minimum is None:  minimum = np.min(data)
        if maximum is None:  maximum = np.max(data)
        if verbose:  print('%.3f < T(all) < %.3f'%(np.min(data), np.max(data))) 
        data = 2*(data - minimum)/(maximum-minimum) - 1.0
        if verbose:  print('%.3f < T(all) < %.3f'%(np.min(data), np.max(data))) 

        # crop maps
        data = data[:,:grid,:grid]

        # get the size and offset depending on the type of dataset
        unique_maps = data.shape[0]
        if   mode=='train':  
            size, offset = int(unique_maps*0.70), int(unique_maps*0.00)
        elif mode=='valid':  
            size, offset = int(unique_maps*0.15), int(unique_maps*0.70)
        elif mode=='test':   
            size, offset = int(unique_maps*0.15), int(unique_maps*0.85)
        elif mode=='all':
            size, offset = int(unique_maps*1.00), int(unique_maps*0.00)
        else:    raise Exception('Wrong name!')

        # randomly shuffle the maps. Instead of 0 1 2 3...999 have a 
        # random permutation. E.g. 5 9 0 29...342
        np.random.seed(seed)
        indexes = np.arange(unique_maps) #only shuffle realizations, not rotations
        np.random.shuffle(indexes)
        indexes = indexes[offset:offset+size] #select indexes of mode

        # keep only the data with the corresponding indexes
        data = data[indexes]

        # define the matrix hosting all data with all rotations/flipping
        # together with the array containing the numbers of each map
        data_all    = np.zeros((size*8, data.shape[1], data.shape[2]), dtype=np.float32)
        numbers_all = np.zeros(size*8, dtype=np.int32)

        # do a loop over all rotations (each is 90 deg)
        total_maps = 0
        for rot in [0,1,2,3]:
            data_rot = np.rot90(data, k=rot, axes=(1,2))

            data_all[total_maps:total_maps+size,:,:] = data_rot
            numbers_all[total_maps:total_maps+size] = indexes
            total_maps += size

            data_all[total_maps:total_maps+size,:,:] = np.flip(data_rot, axis=1)
            numbers_all[total_maps:total_maps+size] = indexes
            total_maps += size
            
        if verbose:
            print('A total of %d maps used'%total_maps)
            print('%.3f < T < %.3f'%(np.min(data), np.max(data)))

        self.size    = data_all.shape[0]
        self.maps    = torch.unsqueeze(torch.tensor(data_all, dtype=torch.float32),1)
        self.numbers = torch.tensor(numbers_all, dtype=torch.int32)

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        return self.maps[idx], self.numbers[idx]


# This routine creates the dataset
def create_dataset(mode, seed, f_images, grid, batch_size, minimum, maximum, 
                   verbose=False):
    data_set    = make_dataset(mode, seed, f_images, grid, minimum, maximum, verbose)
    data_loader = DataLoader(dataset=data_set, batch_size=batch_size, shuffle=True)
    return data_loader

######################################################################################
######################################################################################
######################################################################################

# This class creates the dataset of the fiducial TNG realizations 
class make_dataset_fiducial():

    def __init__(self, mode, seed, f_images, grid, minimum=None, maximum=None,
                 verbose=False):

        # read the data
        data = np.load(f_images) #[number of maps, height, width]

        # normalize maps
        data = np.log10(data)
        if minimum is None:  minimum = np.min(data)
        if maximum is None:  maximum = np.max(data)
        if verbose:  print('%.3f < T(all) < %.3f'%(np.min(data), np.max(data))) 
        data = 2*(data - minimum)/(maximum-minimum) - 1.0
        if verbose:  print('%.3f < T(all) < %.3f'%(np.min(data), np.max(data))) 

        # get all the subimages
        offset_image = 15 #number of pixels to move the frame for a new image
        subimages_dir = (data.shape[1]-grid)//offset_image + 1
        images_tot = subimages_dir*subimages_dir*data.shape[0]
        new_data = np.zeros((images_tot,grid,grid), dtype=np.float32)

        count = 0
        for l in range(data.shape[0]):
            for i in range(subimages_dir):
                x_start, x_end = offset_image*i, offset_image*i+grid
                for j in range(subimages_dir):
                    y_start, y_end = offset_image*j, offset_image*j+grid

                    #print(x_start,x_end)
                    #print(y_start,y_end)
                    new_data[count] = data[l, x_start:x_end, y_start:y_end]
                    count += 1
        print('Using %d maps'%count)


        # get the size and offset depending on the type of dataset
        unique_maps = new_data.shape[0]
        if   mode=='train':  
            size, offset = int(unique_maps*0.70), int(unique_maps*0.00)
        elif mode=='valid':  
            size, offset = int(unique_maps*0.15), int(unique_maps*0.70)
        elif mode=='test':   
            size, offset = int(unique_maps*0.15), int(unique_maps*0.85)
        elif mode=='all':
            size, offset = int(unique_maps*1.00), int(unique_maps*0.00)
        else:    raise Exception('Wrong name!')

        # randomly shuffle the maps. Instead of 0 1 2 3...999 have a 
        # random permutation. E.g. 5 9 0 29...342
        np.random.seed(seed)
        indexes = np.arange(unique_maps) #only shuffle realizations, not rotations
        np.random.shuffle(indexes)
        indexes = indexes[offset:offset+size] #select indexes of mode

        # keep only the data with the corresponding indexes
        data = new_data[indexes]

        # define the matrix hosting all data with all rotations/flipping
        # together with the array containing the numbers of each map
        data_all    = np.zeros((size*8, data.shape[1], data.shape[2]), dtype=np.float32)
        numbers_all = np.zeros(size*8, dtype=np.int32)

        # do a loop over all rotations (each is 90 deg)
        total_maps = 0
        for rot in [0,1,2,3]:
            data_rot = np.rot90(data, k=rot, axes=(1,2))

            data_all[total_maps:total_maps+size,:,:] = data_rot
            numbers_all[total_maps:total_maps+size] = indexes
            total_maps += size

            data_all[total_maps:total_maps+size,:,:] = np.flip(data_rot, axis=1)
            numbers_all[total_maps:total_maps+size] = indexes
            total_maps += size
            
        if verbose:
            print('A total of %d maps used'%total_maps)
            print('%.3f < T < %.3f'%(np.min(data), np.max(data)))

        self.size    = data_all.shape[0]
        self.maps    = torch.unsqueeze(torch.tensor(data_all, dtype=torch.float32),1)
        self.numbers = torch.tensor(numbers_all, dtype=torch.int32)

    def __len__(self):
        return self.size

    def __getitem__(self, idx):
        return self.maps[idx], self.numbers[idx]


# This routine creates the dataset
def create_dataset_fiducial(mode, seed, f_images, grid, batch_size, minimum, maximum, 
                            verbose=False):
    data_set    = make_dataset_fiducial(mode, seed, f_images, grid, minimum, maximum, 
                                        verbose)
    data_loader = DataLoader(dataset=data_set, batch_size=batch_size, shuffle=True)
    return data_loader



# This script computes different quantities for each snapshot of each simulation
from mpi4py import MPI
import numpy as np
import sys,os,h5py,glob
import camel_library as CL


###### MPI DEFINITIONS ###### 
comm   = MPI.COMM_WORLD
nprocs = comm.Get_size()
myrank = comm.Get_rank()


##################################### INPUT ############################################
# CAMEL root folder
root = '/mnt/ceph/users/camels' 

sims = ['IllustrisTNG','SIMBA']#, 'IllustrisTNG']

# Pk parameters
grid    = 512
MAS     = 'CIC'
threads = 1

# halo mass function and baryon fraction parameters
RMmin   = 1e10  #minimum reduced mass (halo mass / Omega_m)
RMmax   = 1e14  #maximum reduced mass (halo mass / Omega_m)
RM_bins = 30

# star-formation rate history parameters
z_min     = 0.0
z_max     = 10.0
sfrh_bins = 10000
BoxSize   = 25.0 #Mpc/h

# stellar mass function parameters
SMmin   = 1e9  #Msun/h
SMmax   = 5e11 #Msun/h
SM_bins = 10

# SO properties
cell_size = 3.0 #Mpc/h maximum radius of a halo (for sorting in a grid)
########################################################################################

# define here an array with all the different realizations
numbers = []
for i in range(1000):  numbers.append('LH_%s'%i)   #LH set
for i in range(1,66):    numbers.append('1P_%s'%i)   #1P set
for i in range(27):    numbers.append('CV_%s'%i)   #CV set
for i in range(4):     numbers.append('EX_%s'%i)   #EX set
numbers = np.array(numbers)

# do a loop over the N-body sims
for sim in ['SIMBA_DM', 'IllustrisTNG_DM']:

    # get the realizations each cpu works on
    indexes      = np.where(np.arange(numbers.shape[0])%nprocs==myrank)[0]
    realizations = numbers[indexes]

    # do a loop over the different realizations
    for i in realizations:
        
        # see if simulation exists
        print(sim,i)
        root_in = '%s/Sims/%s/%s'%(root,sim,i)
        files = glob.glob('%s/*'%root_in)
        if len(files)==0:  continue

        ################# power spectra ###################
        root_out = '%s/Results/Pk/%s/%s'%(root,sim,i)
        if not(os.path.exists(root_out)):  os.makedirs(root_out)
        
        # initial conditions
        snapshot = '%s/Sims/%s/%s/ICs/ics'%(root,sim,i)
        CL.compute_Pk_ICs(snapshot, grid, MAS, threads, [0,1,4,5], root_out)

        # different snapshots
        for snapnum in np.arange(34):
            snapshot = '%s/Sims/%s/%s/snap_%03d.hdf5'%(root,sim,i,snapnum)
            CL.compute_Pk(snapshot, grid, MAS, threads, [0,1,4,5], root_out)
        ###################################################

# do a loop over the different simulations
for sim in sims:

    # get the realizations each cpu works on
    indexes      = np.where(np.arange(numbers.shape[0])%nprocs==myrank)[0]
    realizations = numbers[indexes]

    # do a loop over the different realizations
    for i in realizations:

        # see if simulation exists
        print(sim,i)
        root_in = '%s/Sims/%s/%s'%(root,sim,i)
        files = glob.glob('%s/*'%root_in)
        if len(files)==0:  continue

        ################# power spectra ###################
        root_out = '%s/Results/Pk/%s/%s'%(root,sim,i)
        if not(os.path.exists(root_out)):  os.makedirs(root_out)
        
        # initial conditions
        snapshot = '%s/Sims/%s/%s/ICs/ics'%(root,sim,i)
        CL.compute_Pk_ICs(snapshot, grid, MAS, threads, [0,1,4,5], root_out)

        # different snapshots
        for snapnum in np.arange(34):
            snapshot = '%s/Sims/%s/%s/snap_%03d.hdf5'%(root,sim,i,snapnum)
            CL.compute_Pk(snapshot, grid, MAS, threads, [0,1,4,5], root_out)
            CL.compute_Pk(snapshot, grid, MAS, threads, [0],       root_out)
            CL.compute_Pk(snapshot, grid, MAS, threads, [1],       root_out)
            CL.compute_Pk(snapshot, grid, MAS, threads, [4],       root_out)
            CL.compute_Pk(snapshot, grid, MAS, threads, [5],       root_out)
        ###################################################

        ################ Pk_hydro/Pk_Nbody ################
        root_out = '%s/Results/Pk_ratio/%s/%s'%(root,sim,i)
        if not(os.path.exists(root_out)):  os.makedirs(root_out)
        
        for snapnum in np.arange(34): 
            CL.compute_Pk_ratio(root, sim, i, snapnum, root_out)
        ###################################################

        ########### halo reduced mass function ############
        root_out = '%s/Results/HMF/%s/%s'%(root,sim,i)
        if not(os.path.exists(root_out)):  os.makedirs(root_out)

        # only z=0 for the time being
        snapnum = 33 
        
        snapshot  = '%s/Sims/%s/%s/snap_%03d'%(root,sim,i,snapnum)
        f_subfind = '%s/Sims/%s/%s/fof_subhalo_tab_%03d.hdf5'%(root,sim,i,snapnum)
        CL.halo_mass_function(RMmin, RMmax, RM_bins, f_subfind, snapshot, root_out)
        ###################################################

        ############### baryon fraction ###################
        # only z=0 for the time being
        snapnum   = 33 
        redshift  = 0.0
        snapshot  = '%s/Sims/%s/%s/snap_%03d'%(root,sim,i,snapnum)
        f_subfind = '%s/Sims/%s/%s/fof_subhalo_tab_%03d.hdf5'%(root,sim,i,snapnum)
        f_SO      = '%s/Results/SO/%s/%s/SO_z=%.2f.txt'%(root,sim,i,redshift)

        root_out = '%s/Results/baryon_fraction/%s/%s'%(root,sim,i)
        if not(os.path.exists(root_out)):  os.makedirs(root_out)
        CL.baryon_fraction_FoF(RMmin, RMmax, RM_bins, f_subfind, snapshot, root_out)

        root_out = '%s/Results/baryon_fraction_SO/%s/%s'%(root,sim,i)
        if not(os.path.exists(root_out)):  os.makedirs(root_out)
        CL.baryon_fraction_SO(RMmin, RMmax, RM_bins, f_SO, snapshot, root_out)
        ###################################################

        ######### star-formation rate history #############
        root_out = '%s/Results/SFRH/%s/%s'%(root,sim,i)
        if not(os.path.exists(root_out)):  os.makedirs(root_out)

        f_sfrh = '%s/Sims/%s/%s/sfr.txt'%(root,sim,i)
        CL.star_formation_rate_history(f_sfrh, z_min, z_max, sfrh_bins, BoxSize, root_out)
        ###################################################

        ############ stellar mass function ################
        root_out = '%s/Results/SMF/%s/%s'%(root,sim,i)
        if not(os.path.exists(root_out)):  os.makedirs(root_out)

        f_subfind = '%s/Sims/%s/%s/fof_subhalo_tab_%03d.hdf5'%(root,sim,i,snapnum)
        CL.stellar_mass_function(SMmin, SMmax, SM_bins, f_subfind, root_out)
        ###################################################

        ################# SO properties ###################
        root_out = '%s/Results/SO/%s/%s'%(root,sim,i)
        if not(os.path.exists(root_out)):  os.makedirs(root_out)

        snapshot  = '%s/Sims/%s/%s/snap_%03d.hdf5'%(root,sim,i,snapnum)
        f_subfind = '%s/Sims/%s/%s/fof_subhalo_tab_%03d.hdf5'%(root,sim,i,snapnum)
        CL.SO_properties(snapshot, f_subfind, cell_size, root_out)
        ###################################################

        ################ galaxy properties ################
        root_out  = '%s/Results'%root
        f_subfind = '%s/Sims/%s/%s/fof_subhalo_tab_%03d.hdf5'%(root,sim,i,snapnum)
        CL.properties_vs_SM(SMmin, SMmax, SM_bins, f_subfind, root_out, sim, i)
        ###################################################




from mpi4py import MPI
import numpy as np
import sys,os
import camb

###### MPI DEFINITIONS ######                                    
comm   = MPI.COMM_WORLD
nprocs = comm.Get_size()
myrank = comm.Get_rank()

#################################### INPUT #############################################
root_out = '/simons/scratch/fvillaescusa/CAMELS/Sims/IllustrisTNG'

dimensions = 6
points     = 66
fout       = 'same_seed_params.txt'

# CAMB parameters
Omega_b      = 0.049
h            = 0.6711
ns           = 0.9624
hierarchy    = 'degenerate'
Mnu          = 0.0 #eV
Nnu          = 0   #number of massive neutrinos
Neff         = 3.046
As           = 2.13e-9
tau          = None
Omega_k      = 0.0
pivot_scalar = 0.05
pivot_tensor = 0.05
kmax         = 200.0
k_per_logint = 20
redshifts    = [0]
########################################################################################

# generate the parameters
params = np.zeros((points,dimensions), dtype=np.float64)

# set fiducial values
params[:,0] = 0.3
params[:,1] = 0.8
params[:,2] = 1.0
params[:,3] = 1.0
params[:,4] = 1.0
params[:,5] = 1.0

# set value of sims with different Omega_m
params[:11,  0] = np.linspace(0.1, 0.5, 11)
params[11:22,1] = np.linspace(0.6, 1.0, 11)
params[22:33,2] = np.logspace(np.log10(0.25), np.log10(4.00), 11)
params[33:44,3] = np.logspace(np.log10(0.25), np.log10(4.00), 11)
params[44:55,4] = np.logspace(np.log10(0.50), np.log10(2.00), 11)
params[55:66,5] = np.logspace(np.log10(0.50), np.log10(2.00), 11)
print(params)

# check if the generated parameters are the same as the saved ones
if os.path.exists(fout):  
    params_file = np.loadtxt(fout)
    equal = (np.around(params,5)==params_file) #approximate params with 5 decimals
    if np.any(equal==False):  raise Exception('Cosmological parameters are different!!!')
else:
    np.savetxt(fout, params, fmt='%.5f')

sys.exit()

# get the numbers each cpu will work on
numbers = np.where(np.arange(points)%nprocs==myrank)[0]

# do a loop over all the points in the latin hypercube
for i in numbers:

    # create output folder in case it does not exists
    folder = '%s/%s/'%(root_out,i+1500)
    if not(os.path.exists(folder)):  os.system('mkdir %s'%folder)

    # create initial conditions folder if it does not exists
    folder_ICs = '%s/ICs'%(folder) 
    if not(os.path.exists(folder_ICs)):  os.system('mkdir %s'%folder_ICs)

    # find the values of the cosmological parameters
    Omega_m = params[i,0]
    s8      = params[i,1]
    astro1  = params[i,2]
    astro2  = params[i,3]
    astro3  = params[i,4]
    astro4  = params[i,5]

    g = open('%s/CosmoAstro_params.txt'%folder, 'w')
    g.write('%.5f %.5f %.5f %.5f %.5f %.5f\n'\
            %(Omega_m, s8, astro1, astro2, astro3, astro4))
    g.close()

    print('realization %d'%i)
    print('Omega_m = %.5f'%Omega_m)
    print('s8      = %.5f'%s8)
    print('Astro1  = %.5f'%astro1)
    print('Astro2  = %.5f'%astro2)
    print('Astro3  = %.5f'%astro3)
    print('Astro4  = %.5f'%astro4)

    ##### run CAMB #####
    Omega_c  = Omega_m - Omega_b
    pars     = camb.CAMBparams()

    # set accuracy of the calculation
    pars.set_accuracy(AccuracyBoost=4.0, lSampleBoost=4.0, lAccuracyBoost=4.0, 
                      HighAccuracyDefault=True, DoLateRadTruncation=True)

    # set value of the cosmological parameters
    pars.set_cosmology(H0=h*100.0, ombh2=Omega_b*h**2, omch2=Omega_c*h**2, 
                       mnu=Mnu, omk=Omega_k, neutrino_hierarchy=hierarchy, 
                       num_massive_neutrinos=Nnu, nnu=Neff, tau=tau)
                   
    # set the value of the primordial power spectrum parameters
    pars.InitPower.set_params(As=As, ns=ns, 
                              pivot_scalar=pivot_scalar, pivot_tensor=pivot_tensor)

    # set redshifts, k-range and k-sampling
    pars.set_matter_power(redshifts=redshifts, kmax=kmax, k_per_logint=k_per_logint)

    # compute results
    results = camb.get_results(pars)

    # save parameter values to file
    f = open('%s/CAMB.params'%folder_ICs,'w');  f.write('%s'%pars);  f.close()

    # interpolate to get Pmm, Pcc...etc
    k, zs, Pkmm = results.get_matter_power_spectrum(minkh=2e-5, maxkh=kmax, 
                                                    npoints=400, var1=7, var2=7, 
                                                    have_power_spectra=True, 
                                                    params=None)

    # do a loop over all redshifts
    for j,z in enumerate(zs):
        fout = '%s/Pk_m_z=%.3f.txt'%(folder_ICs,z)
        np.savetxt(fout, np.transpose([k,Pkmm[j,:]]))




    ########################## write 2LPT parameter file #######################
    # parameter file for standard simulations
    a="""
    Nmesh            512
    Nsample          256       
    Box              25000.0   
    FileBase         ics         
    OutputDir        ./
    GlassFile        /simons/scratch/fvillaescusa/hydro_DL/Software/2lpt/GLASS/dummy_glass_CDM_B_64_64.dat   
    GlassTileFac     4         
    Omega            %.4f    
    OmegaLambda      %.4f    
    OmegaBaryon      0.049    
    OmegaDM_2ndSpecies  0.0    
    HubbleParam      0.6711    
    Redshift         127       
    Sigma8           %.4f       
    SphereMode       0         
    WhichSpectrum    2         
    FileWithInputSpectrum   ./Pk_m_z=0.000.txt
    InputSpectrum_UnitLength_in_cm  3.085678e24 
    ShapeGamma       0.201     
    PrimordialIndex  1.0       
    
    Phase_flip          0      
    RayleighSampling    1      
    Seed                67      
    
    NumFilesWrittenInParallel 8  
    UnitLength_in_cm          3.085678e21  
    UnitMass_in_g             1.989e43     
    UnitVelocity_in_cm_per_s  1e5          
    
    WDM_On               0      
    WDM_Vtherm_On        0      
    WDM_PartMass_in_kev  10.0   
    """%(Omega_m, 1.0-Omega_m, s8)

    # save parameters to file                                    
    f = open('%s/2LPT.param'%folder_ICs, 'w');  f.write(a);  f.close()




import numpy as np
import sys,os

##################################### INPUT ########################################
root = '/mnt/ceph/users/camels/Sims'
sims = ['SIMBA', 'IllustrisTNG']
####################################################################################

# get the names of the simulations
names = []
for i in range(1000):  names.append('LH_%d'%i)
for i in range(66):    names.append('1P_%d'%i)
for i in range(27):    names.append('CV_%d'%i)
for i in range(4):     names.append('EX_%d'%i)

# do a loop over the different simulations
for sim in sims:

    # open output file
    g = open('../%s/CosmoAstroSeed_params.txt'%sim, 'w')
    
    # do a loop over all the simulations
    for i in names:

        # read Cosmo+Astro params file
        f1 = '%s/%s/%s/CosmoAstro_params.txt'%(root,sim,i)
        Om,s8,A_SN1,A_AGN1,A_SN2,A_AGN2 = np.loadtxt(f1, unpack=True)

        # read initial random seed
        f2 = '%s/%s/%s/ICs/2LPT.param'%(root,sim,i)
        f = open(f2, 'r')
        for line in f.readlines():
            fields = line.split()
            if len(fields)==0:     continue
            if fields[0]=='Seed':  seed = int(fields[1])
        print('%6s %.5f %.5f %.5f %.5f %.5f %.5f %d'\
              %(i,Om,s8,A_SN1,A_AGN1,A_SN2,A_AGN2,seed))
        g.write('%6s %.5f %.5f %.5f %.5f %.5f %.5f %d\n'\
              %(i,Om,s8,A_SN1,A_AGN1,A_SN2,A_AGN2,seed))

    g.close()



from pylab import *
import numpy as np
from matplotlib.ticker import ScalarFormatter
import matplotlib.gridspec as gridspec
from mpl_toolkits.axes_grid1.inset_locator import mark_inset
from matplotlib.ticker import AutoMinorLocator
from matplotlib.colors import LogNorm
from matplotlib.patches import Ellipse
rcParams["mathtext.fontset"]='cm'

############################### figure ###########################
#fig=figure(figsize=(15,10))     #give dimensions to the figure
##################################################################

################################ INPUT #######################################
#axes range
##############################################################################

############################ subplots ############################
#gs = gridspec.GridSpec(2,1,height_ratios=[5,2])
#ax1=plt.subplot(gs[0])
#ax2=plt.subplot(gs[1])

#make a subplot at a given position and with some given dimensions
#ax2=axes([0.4,0.55,0.25,0.1])

#gs.update(hspace=0.0,wspace=0.4,bottom=0.6,top=1.05)
#subplots_adjust(left=None, bottom=None, right=None, top=None,
#                wspace=0.5, hspace=0.5)

#set minor ticks
#ax1.xaxis.set_minor_locator(AutoMinorLocator(4))
#ax1.yaxis.set_minor_locator(AutoMinorLocator(4))


#ax1.xaxis.set_major_formatter( NullFormatter() )   #unset x label 
#ax1.yaxis.set_major_formatter( NullFormatter() )   #unset y label

# custom xticks 
#ax1.set_xticks([0.25, 0.5, 1.0])
#ax1.set_yticks([0.25, 0.5, 1.0])
#ax1.get_xaxis().set_major_formatter(matplotlib.ticker.ScalarFormatter()) #for log 


#ax1.get_yaxis().set_label_coords(-0.2,0.5)  #align y-axis for multiple plots
##################################################################

##################### special behaviour stuff ####################
#to show error missing error bars in log scale
#ax1.set_yscale('log',nonposy='clip')  #set log scale for the y-axis

#set the x-axis in %f format instead of %e
#ax1.xaxis.set_major_formatter(ScalarFormatter()) 

#set size of ticks
#ax1.tick_params(axis='both', which='major', labelsize=10)
#ax1.tick_params(axis='both', which='minor', labelsize=8)

#set the position of the ylabel 
#ax1.yaxis.set_label_coords(-0.2, 0.4)

#set yticks in scientific notation
#ax1.ticklabel_format(axis='y',style='sci',scilimits=(1,4))

#set the x-axis in %f format instead of %e
#formatter = matplotlib.ticker.FormatStrFormatter('$%.2e$') 
#ax1.yaxis.set_major_formatter(formatter) 

#add two legends in the same plot
#ax5 = ax1.twinx()
#ax5.yaxis.set_major_formatter( NullFormatter() )   #unset y label 
#ax5.legend([p1,p2],['0.0 eV','0.3 eV'],loc=3,prop={'size':14},ncol=1)

#set points to show in the yaxis
#ax1.set_yticks([0,1,2])

#highlight a zoomed region
#mark_inset(ax1, ax2, loc1=2, loc2=4, fc="none",edgecolor='purple')
##################################################################

############################ plot type ###########################
#standard plot
#p1,=ax1.plot(x,y,linestyle='-',marker='None')

#error bar plot with the minimum and maximum values of the error bar interval
#p1=ax1.errorbar(r,xi,yerr=[delta_xi_min,delta_xi_max],lw=1,fmt='o',ms=2,
#               elinewidth=1,capsize=5,linestyle='-') 

#filled area
#p1=ax1.fill_between([x_min,x_max],[1.02,1.02],[0.98,0.98],color='k',alpha=0.2)

#hatch area
#ax1.fill([x_min,x_min,x_max,x_max],[y_min,3.0,3.0,y_min],#color='k',
#         hatch='X',fill=False,alpha=0.5)

#scatter plot
#p1=ax1.scatter(k1,Pk1,c='b',edgecolor='none',s=8,marker='*')

#plot with markers
#pl4,=ax1.plot(ke3,Pk3/Pke3,marker='.',markevery=2,c='r',linestyle='None')

#set size of dashed lines
#ax.plot([0, 1], [0, 1], linestyle='--', dashes=(5, 1)) #length of 5, space of 1

#image plot
#cax = ax1.imshow(densities,cmap=get_cmap('jet'),origin='lower',
#           extent=[x_min, x_max, y_min, y_max],
#           #vmin=min_density,vmax=max_density)
#           norm = LogNorm(vmin=min_density,vmax=max_density))
#cbar = fig.colorbar(cax, ax2, ax=ax1, ticks=[-1, 0, 1]) #in ax2 colorbar of ax1
#cbar.set_label(r"$M_{\rm CSF}\/[h^{-1}M_\odot]$",fontsize=14,labelpad=-50)
#cbar.ax.tick_params(labelsize=10)  #to change size of ticks

#make a polygon
#polygon = Rectangle((0.4,50.0), 20.0, 20.0, edgecolor='purple',lw=0.5,
#                    fill=False)
#ax1.add_artist(polygon)
####################################################################

root = '/mnt/ceph/users/camels/Results/images_1P'
seed = 20

x_min, x_max = 0.0, 25.0
y_min, y_max = 0.0, 25.0

rows = 5
cols = 6

"""
for f, f_out,minimum,maximum,cmap,title in zip(['%s/Images_Mgas_IllustrisTNG.npy'%root,
                                                '%s/Images_Mstar_IllustrisTNG.npy'%root,
                                                '%s/Images_Mcdm_IllustrisTNG.npy'%root,
                                                '%s/Images_HI_IllustrisTNG.npy'%root,
                                                '%s/Images_T_IllustrisTNG.npy'%root,
                                                '%s/Images_Vgas_IllustrisTNG.npy'%root,
                                                '%s/Images_Z_IllustrisTNG.npy'%root,
                                                '%s/Images_ne_IllustrisTNG.npy'%root,
                                                '%s/Images_P_IllustrisTNG.npy'%root],
                                    
                                         ['images_Mgas_IllustrisTNG2.png',
                                          'images_Mstar_IllustrisTNG2.png',
                                          'images_Mcdm_IllustrisTNG2.png',
                                          'images_HI_IllustrisTNG2.png',
                                          'images_T_IllustrisTNG2.png',
                                          'images_Vgas_IllustrisTNG2.png',
                                          'images_Z_IllustrisTNG2.png',
                                          'images_ne_IllustrisTNG2.png',
                                          'images_P_IllustrisTNG2.png'], 
                                         
                            [2e9,  2e8,  5e9,  1e4, 2e3,  30.0,  1e-9, 1e5,  1e2], 
                            [5e13, 1e15, 1e15, 1e13, 2e7, 1000.0, 7e-2, 1e16, 1e11],
                                ['jet', 'nipy_spectral', 'gist_stern',
                                 'magma', 'hot', 'rainbow', 'cubehelix',
                                 'gist_earth', 'terrain'],

                ['Gas density', 'Stellar mass', 'dark matter density',
                 'Neutral hydrogen', 'Gas temperature', 'Gas velocity',
                 'Gas metallicity', 'Electron density', 'Gas pressure']):
"""

root = '/mnt/ceph/users/camels/Results/images_1P'

"""
f = '%s/Images_ne_SIMBA_z=0.00.npy'%root
minimum = 1e5
maximum = 1e15
cmap    = 'gist_earth'
f_out   = 'images_ne_SIMBA.png'
title   = 'Electron density'
"""

"""
f = '%s/Images_P_SIMBA_z=0.00.npy'%root
minimum = 1e3
maximum = 1e9
cmap    = 'terrain'
f_out   = 'images_P_SIMBA.png'
title   = 'Gas pressure'
"""

f = '%s/Images_Z_SIMBA_z=0.00.npy'%root
Zsun    = 0.02
minimum = 1e-7/Zsun
maximum = 4e-2/Zsun
cmap    = 'cubehelix'
f_out   = 'images_Z_SIMBA.png'
title   = 'Gas metallicity'

fig, axs = plt.subplots(rows, cols, figsize=(6.0*1.2, 5.1*1.2))
subplots_adjust(left=None, bottom=None, right=None, top=None, wspace=0.1, hspace=0.1)

for i in range(rows):
    for j in range(cols):
        axs[i,j].tick_params(
            axis='both',          # changes apply to the x-axis
            which='both',      # both major and minor ticks are affected
            bottom=False,      # ticks along the bottom edge are off
            top=False,         # ticks along the top edge are off
            left=False,
            right=False,
            labelleft=False,
            labelbottom=False) # labels along the bottom edge are off

# read data
data = np.load(f)/Zsun
print('%.3e %.3e'%(np.min(data), np.max(data)))
print(data.shape)

#simulations
#1 12 23 34 45 56
#3 14 25 36 47 58
#5 16 27 38 49 60
#7 18 29 40 51 62
#9 20 31 42 53 64

offset = 7 #5
maps_per_sim = 15

indexes = np.array([ 1,  3,  5,  7,  9,
                    12, 14, 16, 18, 20,
                    23, 25, 27, 29, 31,
                    34, 36, 38, 40, 42,
                    45, 47, 49, 51, 53,
                    56, 58, 60, 62, 64])*maps_per_sim + offset 

#indexes = np.ones(30,dtype=np.int32)*3*maps_per_sim+offset
data = data[indexes]

data[np.where(data<minimum)] = minimum

num = 0
for j in range(cols):
    for i in range(rows):
        cax = axs[i,j].imshow(data[num],cmap=get_cmap(cmap),origin='lower',
                    interpolation='bicubic', extent=[x_min, x_max, y_min, y_max],
                              norm = LogNorm(vmin=minimum, vmax=maximum))
        num += 1



axs[0,0].text(10.0,      27.0, r"$\Omega_{\rm m}$", fontsize=16, color='k') 
axs[0,0].text(10.0+27*1, 27.0, r"$\sigma_8$",       fontsize=16, color='k') 
axs[0,0].text(7.0+27*2, 27.0, r"$A_{\rm SN1}$",    fontsize=16, color='k') 
axs[0,0].text(5.0+27*3, 27.0, r"$A_{\rm AGN1}$",   fontsize=16, color='k') 
axs[0,0].text(7.0+27*4, 27.0, r"$A_{\rm SN2}$",    fontsize=16, color='k') 
axs[0,0].text(5.0+27*5, 27.0, r"$A_{\rm AGN2}$",   fontsize=16, color='k') 

axs[0,0].text(-6, 2, "very low", fontsize=13, color='k', rotation='vertical') 
axs[0,0].text(-6, -18, "low", fontsize=13, color='k', rotation='vertical') 
axs[0,0].text(-6, -52, "fiducial", fontsize=13, color='k', rotation='vertical') 
axs[0,0].text(-6, -75, "high", fontsize=13, color='k', rotation='vertical') 
axs[0,0].text(-6, -110, "very high", fontsize=13, color='k', rotation='vertical') 

# colorbar
axb = axes([0.92, 0.11, 0.025, 0.77])
cbar = fig.colorbar(cax, axb, ax=axs[0,0]) #in ax2 colorbar of ax1
cbar.set_label(r"$Z/Z_\odot$",fontsize=14,labelpad=5)
cbar.ax.tick_params(labelsize=10)  #to change size of ticks

#,transform=ax1.transAxes)
#suptitle('%s'%title, size=20, y=0.91)  #for title with several panels
savefig(f_out, bbox_inches='tight', dpi=150)
close(fig)








###############################################################################
#some useful colors:

#'darkseagreen'
#'yellow'
#"hotpink"
#"gold
#"fuchsia"
#"lime"
#"brown"
#"silver"
#"cyan"
#"dodgerblue"
#"darkviolet"
#"magenta"
#"deepskyblue"
#"orchid"
#"aqua"
#"darkorange"
#"coral"
#"lightgreen"
#"salmon"
#"bisque"



import numpy as np
import torch
import sys,os
import torch.backends.cudnn as cudnn
sys.path.append('../')
import data as data
import architecture

#################################### INPUT ##########################################
root_in      = '/mnt/ceph/users/camels'
root_out     = '/mnt/ceph/users/camels/Results/neural_nets/params_2_SFRH'
sim          = 'IllustrisTNG'
bins_SFRH    = 100

# architecture parameters
h1 = 500
h2 = 2000
h3 = 2000
h4 = 100
wd = 1e-8
dr = 0.2

seed         = 1
realizations = 1000
batch_size   = 15
fmodel       = '%s/model.pt'%root_out
fout         = '%s/SFRH_ML.txt'%root_out
#####################################################################################

# get GPU if possible
if torch.cuda.is_available():
    print("CUDA Available")
    device = torch.device('cuda')
else:
    print("CUDA Not Available")
    device = torch.device('cpu')

# get the test dataset
test_loader  = data.create_dataset('test', seed, realizations, root_in, bins_SFRH, 
                                   sim, batch_size, root_out)

# get the number of elements in the test set
size = 0
for x_test, y_test in test_loader:
    size += x_test.shape[0]

# define the array with the results
params_pred = np.zeros((size,bins_SFRH), dtype=np.float32)
params_true = np.zeros((size,bins_SFRH), dtype=np.float32)

# get the parameters of the trained model
model = architecture.model_1hl(6, h1, bins_SFRH, dr)
#model = architecture.model_3hl(6, h1, h2, h3, bins_SFRH, dr)
model.load_state_dict(torch.load(fmodel))
model.to(device=device)

# loop over the different batches and get the prediction
offset = 0
model.eval()
for params_test, SFRH_test in test_loader:
    with torch.no_grad():
        params_test = params_test.to(device)
        SFRH_pred = model(params_test)
        length = params_test.shape[0]
        params_pred[offset:offset+length] = SFRH_pred.cpu().numpy()
        params_true[offset:offset+length] = SFRH_test.numpy()
        offset += length

# compute the rmse
error = np.mean((params_pred - params_true)**2)
print(error)

# load all SFRH and compute mean and std
SFRH = data.read_all_SFRH(root_in, sim, realizations, bins_SFRH, root_out)
SFRH = np.log10(SFRH)
mean = np.mean(SFRH, axis=0, dtype=np.float64)
std  = np.std(SFRH,  axis=0, dtype=np.float64)

# de-normalize
SFRH_pred = 10**(params_pred*std + mean)
SFRH_true = 10**(params_true*std + mean)
log_error = np.mean((np.log10(SFRH_pred) - np.log10(SFRH_true))**2)
print('log_error = %.4f'%np.sqrt(log_error))

# relative error
rel_error = np.mean((np.log10(SFRH_pred) - np.log10(SFRH_true))**2/np.log10(SFRH_true)**2)
print('rel_error = %.4f'%np.sqrt(rel_error))

# save results to file
z = np.linspace(0.0, 7.0, 100)
results = np.zeros((bins_SFRH, 2*size+1), dtype=np.float32)
results[:,0] = z
for i in range(1,size+1):
    results[:,i] = SFRH_pred[i-1]
for i in range(size+1,2*size+1):
    results[:,i] = SFRH_true[i-size-1]
np.savetxt(fout, results)
